# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import atexit
import io
import os
import shutil
import tempfile
from typing import Tuple

DEFAULT_TEMP_FILE_PREFIX = "xcube-"


def new_temp_dir(prefix: str = None, suffix: str = None, dir_path: str = None) -> str:
    """Create new temporary directory using ``tempfile.mkdtemp()``.
    The directory will be removed later when the process ends.

    Args:
        prefix: If not ``None``, the file name will
            begin with that prefix; otherwise, a default prefix is used.
        suffix: If not ``None``, the file name will
            end with that suffix, otherwise there will be no suffix.
        dir_path: If not ``None``, the file will be created
            in that directory; otherwise, a default directory is used.
            The default directory is chosen from a platform-dependent
            list, but the user of the application can control the
            directory location by setting the *TMPDIR*, *TEMP* or *TMP*
            environment variables.

    Returns: The absolute path to the new directory.
    """
    prefix = DEFAULT_TEMP_FILE_PREFIX if prefix is None else prefix
    return remove_dir_later(
        tempfile.mkdtemp(prefix=prefix, suffix=suffix, dir=dir_path)
    )


def new_temp_file(
    prefix: str = None,
    suffix: str = None,
    dir_path: str = None,
    text_mode: bool = False,
) -> tuple[int, str]:
    """Create new temporary file using ``tempfile.mkstemp()``.
    The file will be removed later when the process ends.

    Args:
        prefix: If not ``None``, the file name will begin with that
            prefix; otherwise, a default prefix is used.
        suffix: If not ``None``, the file name will end with that
            suffix, otherwise there will be no suffix.
        dir_path: If not ``None``, the file will be created in that
            directory; otherwise, a default directory is used. The
            default directory is chosen from a platform-dependent list,
            but the user of the application can control the directory
            location by setting the *TMPDIR*, *TEMP* or *TMP*
            environment variables.
        text_mode: If specified and true, the file is opened in text
            mode. Otherwise, (the default) the file is opened in binary
            mode.

    Returns:
        A tuple comprising the file descriptor (an int) and the absolute
        path to the new file.
    """
    prefix = DEFAULT_TEMP_FILE_PREFIX if prefix is None else prefix
    fd, file_path = tempfile.mkstemp(
        prefix=prefix, suffix=suffix, dir=dir_path, text=text_mode
    )
    return fd, remove_file_later(file_path)


def remove_dir_later(dir_path: str) -> str:
    """Remove directory later when the process ends.
    Removal failures are ignored.

    Args:
        dir_path: Path to directory

    Returns:
        The absolute path to the directory.
    """
    dir_path = os.path.abspath(dir_path)

    def _remove_dir_later():
        # noinspection PyBroadException
        try:
            shutil.rmtree(dir_path, ignore_errors=True)
        except BaseException:
            pass

    atexit.register(_remove_dir_later)
    return dir_path


def remove_file_later(file_path: str) -> str:
    """Remove file later when the process ends.
    Removal failures are ignored.

    Args:
        file_path: path to a file

    Returns:
        The absolute path to the file.
    """
    file_path = os.path.abspath(file_path)

    def _remove_file_later():
        # noinspection PyBroadException
        try:
            os.remove(file_path)
        except BaseException:
            pass

    atexit.register(_remove_file_later)
    return file_path



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import xarray as xr

from xcube.server.api import Context
from xcube.webapi.datasets.context import DatasetsContext


def get_datasets_ctx(ctx: Context) -> DatasetsContext:
    return ctx.get_api_ctx("datasets")


def get_dataset(ctx: Context, dataset_id: str | None = None) -> xr.Dataset | None:
    return get_datasets_ctx(ctx).get_dataset(dataset_id) if dataset_id else None



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import datetime
import hashlib
import time
from collections.abc import Mapping
from typing import Any, Dict, List, Optional

from xcube.util.assertions import assert_instance

_CONTENT_LENGTH_DUMMY = -1
_LAST_MODIFIED_DUMMY = str(datetime.datetime.fromtimestamp(time.time(), datetime.UTC))


def list_s3_bucket_v2(
    object_storage: Mapping[str, bytes],
    name: Optional[str] = None,
    delimiter: Optional[str] = None,
    prefix: Optional[str] = None,
    max_keys: Optional[str] = None,
    start_after: Optional[str] = None,
    continuation_token: Optional[str] = None,
    storage_class: Optional[str] = None,
    last_modified: Optional[str] = None,
) -> dict:
    """Implements AWS GET Bucket (List Objects) Version 2
    (https://docs.aws.amazon.com/AmazonS3/latest/API/v2-RESTBucketGET.html)
    for the local filesystem.

    Args:
        object_storage: mapping from path-like keys to byte-objects.
        name: The bucket name.
        delimiter: A delimiter is a character you use to group keys. If
            you specify a prefix, all of the keys that contain the same
            string between the prefix and the first occurrence of the
            delimiter after the prefix are grouped under a single result
            element called CommonPrefixes. If you don't specify the
            prefix parameter, the substring starts at the beginning of
            the key. The keys that are grouped under the CommonPrefixes
            result element are not returned elsewhere in the response.
            Refer to AWS docs for details. No default.
        prefix: Limits the response to keys that begin with the
            specified prefix. You can use prefixes to separate a bucket
            into different groupings of keys. (You can think of using
            prefix to make groups in the same way you'd use a folder in
            a file system.) Refer to AWS docs for details. No default.
        max_keys: Sets the maximum number of keys returned in the
            response body. If you want to retrieve fewer than the
            default 1000 keys, you can add this to your request. The
            response might contain fewer keys, but it never contains
            more. If there are additional keys that satisfy the search
            criteria, but these keys were not returned because max-keys
            was exceeded, the response contains
            <IsTruncated>true</IsTruncated>. To return the additional
            keys, see NextContinuationToken. Refer to AWS docs for
            details. Defaults to 1000.
        start_after: If you want the API to return key names after a
            specific object key in your key space, you can add this
            parameter. Amazon S3 lists objects in UTF-8 character
            encoding in lexicographical order. Refer to AWS docs for
            details. No default.
        continuation_token: When the response to this API call is
            truncated (that is, the IsTruncated response element value
            is true), the response also includes the
            NextContinuationToken element. To list the next set of
            objects, you can use the NextContinuationToken element in
            the next request as the continuation-token. Refer to AWS
            docs for details. No default.
        storage_class: Refer to AWS docs for details. Defaults to
            "STANDARD".
        last_modified: For testing only: always use this value for the
            "LastModified" entry of results

    Returns:
        A dictionary that represents the contents of a
        "ListBucketResult". Refer to AWS docs for details.
    """
    assert_instance(object_storage, collections.abc.Mapping, name="object_storage")
    assert_instance(name, str, name="name")

    max_keys = max_keys or 1000
    start_after = None if continuation_token else start_after
    storage_class = storage_class or "STANDARD"

    contents_list = []
    next_continuation_token = None
    continuation_token_seen = continuation_token is None
    start_key_seen = start_after is None
    common_prefixes_list = []
    common_prefixes_set = set()

    token = 0

    for key in sorted(object_storage.keys()):
        token += 1

        if len(contents_list) == max_keys:
            next_continuation_token = token
            break

        if key == start_after:
            start_key_seen = True
            continue

        if token == continuation_token:
            continuation_token_seen = True

        if not continuation_token_seen or not start_key_seen:
            continue

        if prefix and not key.startswith(prefix):
            continue

        if delimiter:
            index = key.find(delimiter, len(prefix) if prefix else 0)
            if index >= 0:
                key = key[: index + len(delimiter)]
                if key not in common_prefixes_set:
                    common_prefixes_set.add(key)
                    common_prefixes_list.append(key)
                continue

        item = dict(
            Key=key,
            Size=_CONTENT_LENGTH_DUMMY,
            LastModified=last_modified or _LAST_MODIFIED_DUMMY,
            ETag=_str_to_e_tag(key),
            StorageClass=storage_class,
        )
        contents_list.append(item)

    list_bucket_result: dict[str, Any] = dict(
        Name=name,
        Prefix=prefix,
        StartAfter=start_after,
        MaxKeys=max_keys,
        Delimiter=delimiter,
        IsTruncated=next_continuation_token is not None,
        ContinuationToken=continuation_token,
    )
    if next_continuation_token is not None:
        list_bucket_result.update(NextContinuationToken=next_continuation_token)
    if contents_list:
        list_bucket_result.update(Contents=contents_list)
    if common_prefixes_list:
        list_bucket_result.update(
            CommonPrefixes=[dict(Prefix=prefix) for prefix in common_prefixes_list]
        )
    return list_bucket_result


def list_s3_bucket_v1(
    object_storage: Mapping[str, bytes],
    name: str = None,
    delimiter: str = None,
    prefix: str = None,
    max_keys: int = None,
    marker: str = None,
    storage_class: str = None,
    last_modified: str = None,
) -> dict:
    """Implements AWS GET Bucket (List Objects) Version 1
    (https://docs.aws.amazon.com/AmazonS3/latest/API/RESTBucketGET.html)
    for the local filesystem.

    Args:
        object_storage: mapping from path-like keys to byte-objects.
        name: The bucket name, defaults to "s3bucket"
        delimiter: A delimiter is a character you use to group keys. If
            you specify a prefix, all of the keys that contain the same
            string between the prefix and the first occurrence of the
            delimiter after the prefix are grouped under a single result
            element called CommonPrefixes. If you don't specify the
            prefix parameter, the substring starts at the beginning of
            the key. The keys that are grouped under the CommonPrefixes
            result element are not returned elsewhere in the response.
            Refer to AWS docs for details. No default.
        prefix: Limits the response to keys that begin with the
            specified prefix. You can use prefixes to separate a bucket
            into different groupings of keys. (You can think of using
            prefix to make groups in the same way you'd use a folder in
            a file system.) Refer to AWS docs for details. No default.
        max_keys: Sets the maximum number of keys returned in the
            response body. If you want to retrieve fewer than the
            default 1000 keys, you can add this to your request. The
            response might contain fewer keys, but it never contains
            more. If there are additional keys that satisfy the search
            criteria, but these keys were not returned because max-keys
            was exceeded, the response contains
            <IsTruncated>true</IsTruncated>. To return the additional
            keys, see NextMarker. Refer to AWS docs for details.
        marker: Indicates the object key to start with when listing
            objects in a bucket. All objects are listed in the
            dictionary order. Refer to AWS docs for details. No default.
        storage_class: Refer to AWS docs for details. Defaults to
            "STANDARD".
        last_modified: For testing only: always use this value for the
            "LastModified" entry of results

    Returns:
        A dictionary that represents the contents of a
        "ListBucketResult". Refer to AWS docs for details.
    """
    assert_instance(object_storage, collections.abc.Mapping, name="object_storage")
    assert_instance(name, str, name="name")

    max_keys = max_keys or 1000
    storage_class = storage_class or "STANDARD"

    contents_list = []
    is_truncated = False
    next_marker = None
    marker_seen = marker is None
    common_prefixes_list = []
    common_prefixes_set = set()

    for key in sorted(object_storage.keys()):
        if len(contents_list) == max_keys:
            is_truncated = True
            next_marker = key
            break

        if key == marker:
            marker_seen = True

        if not marker_seen:
            continue

        if prefix and not key.startswith(prefix):
            continue

        if delimiter:
            index = key.find(delimiter, len(prefix) if prefix else 0)
            if index >= 0:
                key = key[: index + len(delimiter)]
                if key not in common_prefixes_set:
                    common_prefixes_set.add(key)
                    common_prefixes_list.append(key)
                continue

        item = dict(
            Key=key,
            Size=_CONTENT_LENGTH_DUMMY,
            LastModified=last_modified or _LAST_MODIFIED_DUMMY,
            ETag=_str_to_e_tag(key),
            StorageClass=storage_class,
        )
        contents_list.append(item)

    list_bucket_result: dict[str, Any] = dict(
        Name=name,
        Prefix=prefix,
        Marker=marker,
        MaxKeys=max_keys,
        Delimiter=delimiter,
        IsTruncated=is_truncated,
    )
    if is_truncated:
        list_bucket_result.update(NextMarker=next_marker)
    if contents_list:
        list_bucket_result.update(Contents=contents_list)
    if common_prefixes_list:
        list_bucket_result.update(
            CommonPrefixes=[dict(Prefix=prefix) for prefix in common_prefixes_list]
        )
    return list_bucket_result


def list_bucket_result_to_xml(list_bucket_result):
    return dict_to_xml(
        "ListBucketResult",
        list_bucket_result,
        root_element_attrs=dict(xmlns="http://s3.amazonaws.com/doc/2006-03-01/"),
    )


def dict_to_xml(
    root_element_name: str, content_dict: dict, root_element_attrs: dict = None
) -> str:
    lines = []
    _value_to_xml(
        lines, root_element_name, content_dict, element_attrs=root_element_attrs
    )
    return '<?xml version="1.0" encoding="UTF-8"?>\n' + "\n".join(lines)


def _value_to_xml(
    lines: list[str],
    element_name: str,
    element_value: Any,
    element_attrs: dict = None,
    indent: int = 0,
):
    attrs = ""
    if element_attrs:
        for attr_name, attr_value in element_attrs.items():
            attrs += f' {attr_name}="{_value_to_text(attr_value)}"'
    if element_value is None:
        lines.append(f"{indent * '  '}<{element_name}{attrs}/>")
    elif isinstance(element_value, dict):
        lines.append(f"{indent * '  '}<{element_name}{attrs}>")
        for sub_element_name, sub_element_value in element_value.items():
            _value_to_xml(lines, sub_element_name, sub_element_value, indent=indent + 1)
        lines.append(f"{indent * '  '}</{element_name}>")
    elif isinstance(element_value, list):
        for item in element_value:
            _value_to_xml(lines, element_name, item, indent=indent)
    else:
        lines.append(
            f"{indent * '  '}"
            f"<{element_name}{attrs}>"
            f"{_value_to_text(element_value)}"
            f"</{element_name}>"
        )


def _value_to_text(value) -> str:
    if isinstance(value, bool):
        return "true" if value else "false"
    return str(value)


def _str_to_e_tag(s) -> str:
    return '"' + hashlib.md5(bytes(s, encoding="utf-8")).hexdigest() + '"'



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import itertools
import os
import re
import uuid
import warnings
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import dask.array as da
import dask.array.core as dac
import distributed
import numpy as np

IntTuple = tuple[int, ...]
SliceTuple = tuple[slice, ...]
IntIterable = Iterable[int]
IntTupleIterable = Iterable[IntTuple]
SliceTupleIterable = Iterable[SliceTuple]

_CLUSTER_TAGS_ENV_VAR_NAME = "XCUBE_DASK_CLUSTER_TAGS"
_CLUSTER_ACCOUNT_ENV_VAR_NAME = "XCUBE_DASK_CLUSTER_ACCOUNT"


def compute_array_from_func(
    func: Callable[..., np.ndarray],
    shape: IntTuple,
    chunks: IntTuple,
    dtype: Any,
    name: str = None,
    ctx_arg_names: Sequence[str] = None,
    args: Sequence[Any] = None,
    kwargs: Mapping[str, Any] = None,
) -> da.Array:
    """Compute a dask array using the provided user function
    *func*, *shape*, and chunking *chunks*.

    The user function is expected to output the array's data
    blocks using arguments specified by *ctx_arg_names*, *args*,
    and *kwargs* and is expected to return a numpy array.

    You can request array and current block context information
    by specifying the optional *ctx_arg_names* keyword argument
    that is a sequence of names of special arguments passed to
    *user_func*. The following are available:

    * ``shape``: The array's shape. A tuple of ints.
    * ``chunks``: The array's chunks. A tuple of tuple of ints.
    * ``dtype``: The array's numpy data type.
    * ``name``: The array's name. A string or ``None``.
    * ``block_id``: The block's unique ID. An integer number
        ranging from zero to number of blocks minus one.
    * ``block_index``: The block's index as a tuple of ints.
    * ``block_shape``: The block's shape as a tuple of ints.
    * ``block_slices``: The block's shape as a tuple of int pair tuples.

    Args:
        func: User function that is called for each block of the
            array using arguments specified by *ctx_arg_names*,
            *args*, and *kwargs*. It must return a numpy array of
            shape "block_shape" and type *dtype*.
        shape: The array's shape. A tuple of sizes for each
            array dimension.
        chunks: The array's chunking. A tuple of chunk sizes for
            each array dimension. Must be of same length as *shape*.
        dtype: The array's numpy data type.
        name: The array's name.
        ctx_arg_names: Sequence names of arguments that are passed
            before *args* and *kwargs* to the user function.
        args: Arguments passed to the user function.
        kwargs: Keyword-arguments passed to the user function.

    Returns: A chunked dask array.
    """
    ctx_arg_names = ctx_arg_names or []
    args = args or []
    kwargs = kwargs or {}

    chunk_sizes = tuple(get_chunk_sizes(shape, chunks))
    chunk_counts = tuple(get_chunk_counts(shape, chunks))
    block_indexes, block_shapes, block_slices = get_block_iterators(chunk_sizes)

    ctx_values = dict(
        shape=tuple(shape),
        chunks=chunk_sizes,
        dtype=dtype,
        name=name,
    )

    blocks = _NestedList(shape=chunk_counts)
    block_id = 0
    for chunk_index, chunk_shape, block_slices in zip(
        block_indexes, block_shapes, block_slices
    ):
        ctx_values.update(
            block_id=block_id,
            block_index=tuple(chunk_index),
            block_shape=tuple(chunk_shape),
            block_slices=tuple(
                (chunk_slice.start, chunk_slice.stop) for chunk_slice in block_slices
            ),
        )
        ctx_args = [ctx_values[ctx_arg_name] for ctx_arg_name in ctx_arg_names]
        block_id += 1

        # We use our own name here, because dac.from_func() tokenizes args which for some reason takes forever
        block = dac.from_func(
            func,
            shape=chunk_shape,
            dtype=dtype,
            name=f"rectify_{name}-{uuid.uuid4()}",
            args=(*ctx_args, *args),
            kwargs=kwargs,
        )

        blocks[chunk_index] = block

    return da.block(blocks.data)


def get_block_iterators(
    chunk_sizes: IntTupleIterable,
) -> tuple[IntTupleIterable, IntTupleIterable, SliceTupleIterable]:
    chunk_sizes = tuple(chunk_sizes)
    chunk_slices_tuples = get_chunk_slice_tuples(chunk_sizes)
    chunk_ranges = get_chunk_ranges(chunk_sizes)
    block_indexes = itertools.product(*chunk_ranges)
    block_shapes = itertools.product(*chunk_sizes)
    block_slices = itertools.product(*chunk_slices_tuples)
    return block_indexes, block_shapes, block_slices


def get_chunk_sizes(shape: IntTuple, chunks: IntTuple) -> IntTupleIterable:
    for s, c in zip(shape, chunks):
        n = s // c
        if n * c < s:
            yield (c,) * n + (s % c,)
        else:
            yield (c,) * n


def get_chunk_counts(shape: IntTuple, chunks: IntTuple) -> Iterable[int]:
    for s, c in zip(shape, chunks):
        yield (s + c - 1) // c


def get_chunk_ranges(chunk_size_tuples: IntTupleIterable) -> Iterable[range]:
    return (range(len(chunk_size_tuple)) for chunk_size_tuple in chunk_size_tuples)


def get_chunk_slice_tuples(chunk_size_tuples: IntTupleIterable) -> SliceTupleIterable:
    return (
        tuple(get_chunk_slices(chunk_size_tuple))
        for chunk_size_tuple in chunk_size_tuples
    )


def get_chunk_slices(chunk_sizes: Sequence[int]) -> Iterable[slice]:
    stop = 0
    for i in range(len(chunk_sizes)):
        start = stop
        stop = start + chunk_sizes[i]
        yield slice(start, stop)


def new_cluster(
    provider: str = "coiled",
    name: Optional[str] = None,
    software: Optional[str] = None,
    n_workers: int = 4,
    resource_tags: Optional[dict[str, str]] = None,
    account: str = None,
    region: str = "eu-central-1",
    **kwargs,
) -> distributed.deploy.Cluster:
    """Create a new Dask cluster.

    Cloud resource tags can be specified in an environment variable
    XCUBE_DASK_CLUSTER_TAGS in the format
    ``tag_1=value_1:tag_2=value_2:...:tag_n=value_n``. In case of
    conflicts, tags specified in ``resource_tags`` will override tags
    specified by the environment variable.

    The cluster provider account name can be specified in an environment
    variable ``XCUBE_DASK_CLUSTER_ACCOUNT``. If the ``account`` argument is
    given to ``new_cluster``, it will override the value from the environment
    variable.

    Args:
        provider: identifier of the provider to use. Currently, only
            'coiled' is supported.
        name: name to use as an identifier for the cluster
        software: identifier for the software environment to be used.
        n_workers: number of workers in the cluster
        resource_tags: tags to apply to the cloud resources forming the
            cluster
        account: cluster provider account name
        **kwargs: further named arguments will be passed on to the
            cluster creation function
        region: default region where workers of the cluster will be
            deployed set to eu-central-1
    """

    if resource_tags is None:
        resource_tags = {}
    if _CLUSTER_ACCOUNT_ENV_VAR_NAME in os.environ:
        account_from_env_var = os.environ[_CLUSTER_ACCOUNT_ENV_VAR_NAME]
    else:
        account_from_env_var = None
        warnings.warn(
            f"Environment variable {_CLUSTER_ACCOUNT_ENV_VAR_NAME}"
            f" not set; cluster account name may be incorrect."
        )

    cluster_account = (
        account
        if account is not None
        else account_from_env_var if account_from_env_var is not None else "bc"
    )

    if provider == "coiled":
        try:
            import coiled
        except ImportError as e:
            raise ImportError(
                f"provider 'coiled' requires package'coiled' to be installed"
            ) from e
        if software is None and "JUPYTER_IMAGE" in os.environ:
            # If the JUPYTER_IMAGE environment variable is set, we're
            # presumably in a Z2JH deployment and can base a
            # Coiled environment on the same image.
            # First we construct an identifier from the user image specifier.
            current_image = os.environ["JUPYTER_IMAGE"]
            software = re.sub(
                "[:.]",
                "-",
                re.search(r"/([^/]+)$", current_image).group(1),
            )
            # If the referenced software environment doesn't exist yet as a
            # Coiled environment, create it from the currently used image.
            available_environments = coiled.list_software_environments(
                account=account
            ).keys()
            if software not in available_environments:
                coiled.create_software_environment(
                    name=software, container=current_image
                )

        # If software is (still) None, Coiled will try to mirror the current
        # environment automagically.
        coiled_params = dict(
            n_workers=n_workers,
            environ=None,
            tags=_collate_cluster_resource_tags(resource_tags),
            account=cluster_account,
            name=name,
            software=software,
            use_best_zone=True,
            compute_purchase_option="spot_with_fallback",
            shutdown_on_close=True,
            region=region,
        )
        coiled_params.update(kwargs)

        return coiled.Cluster(**coiled_params)

    raise NotImplementedError(f"Unknown provider {provider!r}")


def _collate_cluster_resource_tags(extra_tags: dict[str, str]) -> dict[str, str]:
    fallback_tags = {
        "cost-center": "unknown",
        "environment": "dev",
        "creator": "auto",
        "purpose": "xcube dask cluster",
        "user": (
            os.environ.get("JUPYTERHUB_USER")  # JupyterHub
            or os.environ.get("USER")  # Unixes
            or os.environ.get("USERNAME")  # Windows
            or os.getlogin()
            or ""
        ),
    }
    if _CLUSTER_TAGS_ENV_VAR_NAME in os.environ:
        kvps = os.environ[_CLUSTER_TAGS_ENV_VAR_NAME].split(":")
        env_var_tags = {
            (parts := kvp.split("=", maxsplit=1))[0]: parts[1] for kvp in kvps
        }
    else:
        warnings.warn(
            f"Environment variable {_CLUSTER_TAGS_ENV_VAR_NAME}"
            f" not set; cluster resource tags may be missing."
        )
        env_var_tags = {}
    return fallback_tags | env_var_tags | extra_tags


class _NestedList:
    """Utility class whose instances are used as input to dask.block()."""

    def __init__(self, shape: Sequence[int], fill_value: Any = None):
        self._shape = tuple(shape)
        self._data = self._new_data(shape, len(shape), fill_value, 0)

    @classmethod
    def _new_data(
        cls, shape: Sequence[int], ndim: int, fill_value: Any, dim: int
    ) -> Union[list[list], list[Any]]:
        return [
            (
                cls._new_data(shape, ndim, fill_value, dim + 1)
                if dim < ndim - 1
                else fill_value
            )
            for _ in range(shape[dim])
        ]

    @property
    def shape(self) -> tuple[int, ...]:
        return self._shape

    @property
    def data(self) -> Union[list[list], list[Any]]:
        return self._data

    def __len__(self) -> int:
        return len(self._data)

    def __setitem__(self, index: Union[int, slice, tuple], value: Any):
        data = self._data
        if isinstance(index, tuple):
            n = len(index)
            for i in range(n - 1):
                data = data[index[i]]
            data[index[n - 1]] = value
        else:
            data[index] = value

    def __getitem__(self, index: Union[int, slice, tuple]) -> Any:
        data = self._data
        if isinstance(index, tuple):
            n = len(index)
            for i in range(n - 1):
                data = data[index[i]]
            return data[index[n - 1]]
        else:
            return data[index]



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import Api

api = Api("styles", description="xcube Styles API")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

# noinspection PyUnresolvedReferences
from .api import api



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from collections.abc import Mapping, Sequence
from typing import Optional, Union

import dask.array as da
import numba as nb
import numpy as np
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.select import select_spatial_subset
from xcube.util.dask import compute_array_from_func

from .cf import complete_resampled_dataset

_INTERPOLATIONS = {"nearest": 0, "triangular": 1, "bilinear": 2}


def rectify_dataset(
    source_ds: xr.Dataset,
    /,
    source_gm: Optional[GridMapping] = None,
    target_gm: Optional[GridMapping] = None,
    ref_ds: Optional[xr.Dataset] = None,
    var_names: Optional[Union[str, Sequence[str]]] = None,
    encode_cf: bool = True,
    gm_name: Optional[str] = None,
    tile_size: Optional[Union[int, tuple[int, int]]] = None,
    is_j_axis_up: Optional[bool] = None,
    output_ij_names: Optional[tuple[str, str]] = None,
    compute_subset: bool = True,
    uv_delta: float = 1e-3,
    interpolation: Optional[str] = None,
    xy_var_names: Optional[tuple[str, str]] = None,
) -> Optional[xr.Dataset]:
    """Reproject dataset *source_ds* using its per-pixel
    x,y coordinates or the given *source_gm*.

    The function expects *source_ds* or the given
    *source_gm* to have either one- or two-dimensional
    coordinate variables that provide spatial x,y coordinates
    for every data variable with the same spatial dimensions.

    For example, a dataset may comprise variables with
    spatial dimensions ``var(..., y_dim, x_dim)``,
    then the function expects coordinates to be provided
    in two forms:

    1. One-dimensional ``x_var(x_dim)``
       and ``y_var(y_dim)`` (coordinate) variables.
    2. Two-dimensional ``x_var(y_dim, x_dim)``
       and ``y_var(y_dim, x_dim)`` (coordinate) variables.

    If *target_gm* is given, and it defines a tile size,
    or *tile_size* is given and the number of tiles is
    greater than one in the output's x- or y-direction, then the
    returned dataset will be composed of lazy, chunked dask
    arrays. Otherwise, the returned dataset will be composed
    of ordinary numpy arrays.

    New in 1.6: If *target_ds* is given, its coordinate
    variables are copied by reference into the returned
    dataset.

    Args:
        source_ds: Source dataset.
        source_gm: Source dataset grid mapping.
        target_gm: Optional target geometry. If not given, output
            geometry will be computed to spatially fit *dataset* and to
            retain its spatial resolution.
        ref_ds: An optional dataset that provides the
            target grid mapping if *target_gm* is not provided.
            If *ref_ds* is given, its coordinate variables are copied
            by reference into the returned dataset.
        var_names: Optional variable name or sequence of variable names.
        encode_cf: Whether to encode the target grid mapping into the
            resampled dataset in a CF-compliant way. Defaults to
            ``True``.
        gm_name: Name for the grid mapping variable. Defaults to "crs".
            Used only if *encode_cf* is ``True``.
        tile_size: Optional tile size for the output.
        is_j_axis_up: Whether y coordinates are increasing with positive
            image j axis.
        output_ij_names: If given, a tuple of variable names in which to
            store the computed source pixel coordinates in the returned
            output.
        compute_subset: Whether to compute a spatial subset from
            *source_ds* using the boundary of the target grid mapping.
            If set, the function may return ``None`` in case there is no
            overlap.
        uv_delta: A normalized value that is used to determine whether
            x,y coordinates in the output are contained in the triangles
            defined by the input x,y coordinates. The higher this value,
            the more inaccurate the rectification will be.
        interpolation: Interpolation method for computing output pixels.
            If given, must be "nearest", "triangular", or "bilinear".
            The default is "nearest". The "triangular" interpolation is
            performed between 3 and "bilinear" between 4 adjacent source
            pixels. Both are applied only to variables of
            floating point type. If you need to interpolate between
            integer data you should cast it to float first.
        xy_var_names: Deprecated. No longer used since 1.0.0,
            no replacement.

    Returns:
        A reprojected dataset, or None if the requested output does not
        intersect with *dataset*.
    """
    if xy_var_names:
        warnings.warn(
            "argument 'xy_var_names' has been deprecated in 1.4.2"
            " and may be removed anytime.",
            category=DeprecationWarning,
        )

    if source_gm is None:
        source_gm = GridMapping.from_dataset(source_ds)

    src_attrs = dict(source_ds.attrs)

    if target_gm is None and ref_ds is not None:
        target_gm = GridMapping.from_dataset(ref_ds)

    if target_gm is None:
        target_gm = source_gm.to_regular(tile_size=tile_size)
    elif compute_subset:
        source_ds_subset = select_spatial_subset(
            source_ds,
            xy_bbox=target_gm.xy_bbox,
            ij_border=1,
            xy_border=0.5 * (target_gm.x_res + target_gm.y_res),
            grid_mapping=source_gm,
        )
        if source_ds_subset is None:
            return None
        if source_ds_subset is not source_ds:
            source_gm = GridMapping.from_dataset(source_ds_subset)
            source_ds = source_ds_subset

    if tile_size is not None or is_j_axis_up is not None:
        target_gm = target_gm.derive(tile_size=tile_size, is_j_axis_up=is_j_axis_up)

    src_vars = _select_variables(source_ds, source_gm, var_names)

    interpolation_mode = _INTERPOLATIONS.get(interpolation or "nearest")
    if interpolation_mode is None:
        raise ValueError(f"invalid interpolation: {interpolation!r}")

    if target_gm.is_tiled:
        compute_dst_src_ij_images = _compute_ij_images_xarray_dask
        compute_dst_var_image = _compute_var_image_xarray_dask
    else:
        compute_dst_src_ij_images = _compute_ij_images_xarray_numpy
        compute_dst_var_image = _compute_var_image_xarray_numpy

    dst_src_ij_array = compute_dst_src_ij_images(source_gm, target_gm, uv_delta)

    dst_x_dim, dst_y_dim = target_gm.xy_dim_names
    dst_dims = dst_y_dim, dst_x_dim
    dst_ds_coords = target_gm.to_coords()
    dst_vars = dict()
    for src_var_name, src_var in src_vars.items():
        dst_var_dims = src_var.dims[0:-2] + dst_dims
        dst_var_coords = {
            d: src_var.coords[d] for d in dst_var_dims if d in src_var.coords
        }
        # noinspection PyTypeChecker
        dst_var_coords.update(
            {d: dst_ds_coords[d] for d in dst_var_dims if d in dst_ds_coords}
        )
        dst_var_array = compute_dst_var_image(
            src_var,
            dst_src_ij_array,
            fill_value=np.nan,
            interpolation=interpolation_mode,
        )
        dst_var = xr.DataArray(
            dst_var_array,
            dims=dst_var_dims,
            coords=dst_var_coords,
            attrs=src_var.attrs,
        )
        dst_vars[src_var_name] = dst_var

    if output_ij_names:
        output_i_name, output_j_name = output_ij_names
        dst_ij_coords = {d: dst_ds_coords[d] for d in dst_dims if d in dst_ds_coords}
        dst_vars[output_i_name] = xr.DataArray(
            dst_src_ij_array[0], dims=dst_dims, coords=dst_ij_coords
        )
        dst_vars[output_j_name] = xr.DataArray(
            dst_src_ij_array[1], dims=dst_dims, coords=dst_ij_coords
        )

    return complete_resampled_dataset(
        encode_cf,
        xr.Dataset(dst_vars, coords=dst_ds_coords, attrs=src_attrs),
        target_gm,
        gm_name,
        ref_ds.coords if ref_ds else None,
    )


def _select_variables(
    source_ds: xr.Dataset,
    source_gm: GridMapping,
    var_names: Union[None, str, Sequence[str]],
) -> Mapping[str, xr.DataArray]:
    """Select variables from *dataset*.

    Args:
        source_ds: Source dataset.
        source_gm: Optional dataset geo-coding.
        var_names: Optional variable name or sequence of variable names.

    Returns:
        The selected variables as a variable name to ``xr.DataArray``
        mapping
    """
    spatial_var_names = source_gm.xy_var_names
    spatial_shape = tuple(reversed(source_gm.size))
    spatial_dims = tuple(reversed(source_gm.xy_dim_names))
    if var_names is None:
        var_names = [
            var_name
            for var_name, var in source_ds.data_vars.items()
            if var_name not in spatial_var_names
            and _is_2d_spatial_var(var, spatial_shape, spatial_dims)
        ]
    elif isinstance(var_names, str):
        var_names = (var_names,)
    elif len(var_names) == 0:
        raise ValueError(f"empty var_names")
    src_vars = {}
    for var_name in var_names:
        src_var = source_ds[var_name]
        if not _is_2d_spatial_var(src_var, spatial_shape, spatial_dims):
            raise ValueError(
                f"cannot rectify variable {var_name!r}"
                f" as its shape or dimensions"
                f" do not match those of {spatial_var_names[0]!r}"
                f" and {spatial_var_names[1]!r}"
            )
        src_vars[var_name] = src_var
    return src_vars


def _is_2d_spatial_var(var: xr.DataArray, shape, dims) -> bool:
    return var.ndim >= 2 and var.shape[-2:] == shape and var.dims[-2:] == dims


def _compute_ij_images_xarray_numpy(
    src_geo_coding: GridMapping, output_geom: GridMapping, uv_delta: float
) -> np.ndarray:
    """Compute numpy.ndarray destination image with source
    pixel i,j coords from xarray.DataArray x,y sources.
    """
    dst_width = output_geom.width
    dst_height = output_geom.height
    dst_shape = 2, dst_height, dst_width
    dst_src_ij_images = np.full(dst_shape, np.nan, dtype=np.float64)
    dst_x_offset = output_geom.x_min
    dst_y_offset = output_geom.y_min if output_geom.is_j_axis_up else output_geom.y_max
    dst_x_scale = output_geom.x_res
    dst_y_scale = output_geom.y_res if output_geom.is_j_axis_up else -output_geom.y_res
    x_values, y_values = src_geo_coding.xy_coords.values
    _compute_ij_images_numpy_parallel(
        x_values,
        y_values,
        0,
        0,
        dst_src_ij_images,
        dst_x_offset,
        dst_y_offset,
        dst_x_scale,
        dst_y_scale,
        uv_delta,
    )
    return dst_src_ij_images


def _compute_ij_images_xarray_dask(
    src_geo_coding: GridMapping, output_geom: GridMapping, uv_delta: float
) -> da.Array:
    """Compute dask.array.Array destination image
    with source pixel i,j coords from xarray.DataArray x,y sources.
    """
    dst_width = output_geom.width
    dst_height = output_geom.height
    dst_tile_width = output_geom.tile_width
    dst_tile_height = output_geom.tile_height
    dst_var_shape = 2, dst_height, dst_width
    dst_var_chunks = 2, dst_tile_height, dst_tile_width

    dst_x_min, dst_y_min, dst_x_max, dst_y_max = output_geom.xy_bbox
    dst_x_res, dst_y_res = output_geom.xy_res
    dst_is_j_axis_up = output_geom.is_j_axis_up

    # Compute an empirical xy_border as a function of the
    # number of tiles, because the more tiles we have
    # the smaller the destination xy-bboxes and the higher
    # the risk to not find any source ij-bbox for a given xy-bbox.
    # xy_border will not be larger than half of the
    # coverage of a tile.
    #
    num_tiles_x = dst_width / dst_tile_width
    num_tiles_y = dst_height / dst_tile_height
    xy_border = min(
        min(2 * num_tiles_x * output_geom.x_res, 2 * num_tiles_y * output_geom.y_res),
        min(0.5 * (dst_x_max - dst_x_min), 0.5 * (dst_y_max - dst_y_min)),
    )

    dst_xy_bboxes = output_geom.xy_bboxes
    src_ij_bboxes = src_geo_coding.ij_bboxes_from_xy_bboxes(
        dst_xy_bboxes, xy_border=xy_border, ij_border=1
    )

    return compute_array_from_func(
        _compute_ij_images_xarray_dask_block,
        dst_var_shape,
        dst_var_chunks,
        np.float64,
        ctx_arg_names=[
            "dtype",
            "block_id",
            "block_shape",
            "block_slices",
        ],
        args=(
            src_geo_coding.xy_coords,
            src_ij_bboxes,
            dst_x_min,
            dst_y_min,
            dst_y_max,
            dst_x_res,
            dst_y_res,
            dst_is_j_axis_up,
            uv_delta,
        ),
        name="ij_pixels",
    )


def _compute_ij_images_xarray_dask_block(
    dtype: np.dtype,
    block_id: int,
    block_shape: tuple[int, int],
    block_slices: tuple[tuple[int, int], tuple[int, int], tuple[int, int]],
    src_xy_coords: xr.DataArray,
    src_ij_bboxes: np.ndarray,
    dst_x_min: float,
    dst_y_min: float,
    dst_y_max: float,
    dst_x_res: float,
    dst_y_res: float,
    dst_is_j_axis_up: bool,
    uv_delta: float,
) -> np.ndarray:
    """Compute dask.array.Array destination block with source
    pixel i,j coords from xarray.DataArray x,y sources.
    """
    dst_src_ij_block = np.full(block_shape, np.nan, dtype=dtype)
    _, (dst_y_slice_start, _), (dst_x_slice_start, _) = block_slices
    src_ij_bbox = src_ij_bboxes[block_id]
    src_i_min, src_j_min, src_i_max, src_j_max = src_ij_bbox
    if src_i_min == -1:
        return dst_src_ij_block
    src_xy_values = src_xy_coords[
        :, src_j_min : src_j_max + 1, src_i_min : src_i_max + 1
    ].values
    src_x_values = src_xy_values[0]
    src_y_values = src_xy_values[1]
    dst_x_offset = dst_x_min + dst_x_slice_start * dst_x_res
    if dst_is_j_axis_up:
        dst_y_offset = dst_y_min + dst_y_slice_start * dst_y_res
    else:
        dst_y_offset = dst_y_max - dst_y_slice_start * dst_y_res
    _compute_ij_images_numpy_sequential(
        src_x_values,
        src_y_values,
        src_i_min,
        src_j_min,
        dst_src_ij_block,
        dst_x_offset,
        dst_y_offset,
        dst_x_res,
        dst_y_res if dst_is_j_axis_up else -dst_y_res,
        uv_delta,
    )
    return dst_src_ij_block


@nb.njit(nogil=True, parallel=True, cache=True)
def _compute_ij_images_numpy_parallel(
    src_x_image: np.ndarray,
    src_y_image: np.ndarray,
    src_i_min: int,
    src_j_min: int,
    dst_src_ij_images: np.ndarray,
    dst_x_offset: float,
    dst_y_offset: float,
    dst_x_scale: float,
    dst_y_scale: float,
    uv_delta: float,
):
    """Compute numpy.ndarray destination image with source
    pixel i,j coords from numpy.ndarray x,y sources in
    parallel mode.
    """
    src_height = src_x_image.shape[-2]
    dst_src_ij_images[:, :, :] = np.nan
    for src_j0 in nb.prange(src_height - 1):
        _compute_ij_images_for_source_line(
            src_j0,
            src_x_image,
            src_y_image,
            src_i_min,
            src_j_min,
            dst_src_ij_images,
            dst_x_offset,
            dst_y_offset,
            dst_x_scale,
            dst_y_scale,
            uv_delta,
        )


# Extra dask version, because if we use parallel=True
# and nb.prange, we end up in infinite JIT compilation :(
@nb.njit(nogil=True, cache=True)
def _compute_ij_images_numpy_sequential(
    src_x_image: np.ndarray,
    src_y_image: np.ndarray,
    src_i_min: int,
    src_j_min: int,
    dst_src_ij_images: np.ndarray,
    dst_x_offset: float,
    dst_y_offset: float,
    dst_x_scale: float,
    dst_y_scale: float,
    uv_delta: float,
):
    """Compute numpy.ndarray destination image with source pixel i,j coords
    from numpy.ndarray x,y sources NOT in parallel mode.
    """
    src_height = src_x_image.shape[-2]
    dst_src_ij_images[:, :, :] = np.nan
    for src_j0 in range(src_height - 1):
        _compute_ij_images_for_source_line(
            src_j0,
            src_x_image,
            src_y_image,
            src_i_min,
            src_j_min,
            dst_src_ij_images,
            dst_x_offset,
            dst_y_offset,
            dst_x_scale,
            dst_y_scale,
            uv_delta,
        )


@nb.njit(nogil=True, cache=True)
def _compute_ij_images_for_source_line(
    src_j0: int,
    src_x_image: np.ndarray,
    src_y_image: np.ndarray,
    src_i_min: int,
    src_j_min: int,
    dst_src_ij_images: np.ndarray,
    dst_x_offset: float,
    dst_y_offset: float,
    dst_x_scale: float,
    dst_y_scale: float,
    uv_delta: float,
):
    """Compute numpy.ndarray destination image with source
    pixel i,j coords from a numpy.ndarray x,y source line.
    """
    src_width = src_x_image.shape[-1]

    dst_width = dst_src_ij_images.shape[-1]
    dst_height = dst_src_ij_images.shape[-2]

    dst_px = np.zeros(4, dtype=src_x_image.dtype)
    dst_py = np.zeros(4, dtype=src_y_image.dtype)

    u_min = v_min = -uv_delta
    uv_max = 1.0 + 2 * uv_delta

    for src_i0 in range(src_width - 1):
        src_i1 = src_i0 + 1
        src_j1 = src_j0 + 1

        dst_px[0] = dst_p0x = src_x_image[src_j0, src_i0]
        dst_px[1] = dst_p1x = src_x_image[src_j0, src_i1]
        dst_px[2] = dst_p2x = src_x_image[src_j1, src_i0]
        dst_px[3] = dst_p3x = src_x_image[src_j1, src_i1]

        dst_py[0] = dst_p0y = src_y_image[src_j0, src_i0]
        dst_py[1] = dst_p1y = src_y_image[src_j0, src_i1]
        dst_py[2] = dst_p2y = src_y_image[src_j1, src_i0]
        dst_py[3] = dst_p3y = src_y_image[src_j1, src_i1]

        dst_pi = np.floor((dst_px - dst_x_offset) / dst_x_scale).astype(np.int64)
        dst_pj = np.floor((dst_py - dst_y_offset) / dst_y_scale).astype(np.int64)

        dst_i_min = np.min(dst_pi)
        dst_i_max = np.max(dst_pi)
        dst_j_min = np.min(dst_pj)
        dst_j_max = np.max(dst_pj)

        if (
            dst_i_max < 0
            or dst_j_max < 0
            or dst_i_min >= dst_width
            or dst_j_min >= dst_height
        ):
            continue

        if dst_i_min < 0:
            dst_i_min = 0

        if dst_i_max >= dst_width:
            dst_i_max = dst_width - 1

        if dst_j_min < 0:
            dst_j_min = 0

        if dst_j_max >= dst_height:
            dst_j_max = dst_height - 1

        # u from p0 right to p1, v from p0 down to p2
        # noinspection PyTypeChecker
        det_a = _fdet(dst_p0x, dst_p0y, dst_p1x, dst_p1y, dst_p2x, dst_p2y)
        if np.isnan(det_a):
            det_a = 0.0

        # u from p3 left to p2, v from p3 up to p1
        # noinspection PyTypeChecker
        det_b = _fdet(dst_p3x, dst_p3y, dst_p2x, dst_p2y, dst_p1x, dst_p1y)
        if np.isnan(det_b):
            det_b = 0.0

        if det_a == 0.0 and det_b == 0.0:
            # Both the triangles do not exist.
            continue

        for dst_j in range(dst_j_min, dst_j_max + 1):
            dst_y = dst_y_offset + (dst_j + 0.5) * dst_y_scale
            for dst_i in range(dst_i_min, dst_i_max + 1):
                sentinel = dst_src_ij_images[0, dst_j, dst_i]
                if not np.isnan(sentinel):
                    # If we have a source pixel in dst_i, dst_j already,
                    # there is no need to compute another one.
                    # One is as good as the other.
                    continue

                dst_x = dst_x_offset + (dst_i + 0.5) * dst_x_scale

                src_i = src_j = -1

                if det_a != 0.0:
                    # noinspection PyTypeChecker
                    u = _fu(dst_x, dst_y, dst_p0x, dst_p0y, dst_p2x, dst_p2y) / det_a
                    # noinspection PyTypeChecker
                    v = _fv(dst_x, dst_y, dst_p0x, dst_p0y, dst_p1x, dst_p1y) / det_a
                    if u >= u_min and v >= v_min and u + v <= uv_max:
                        src_i = src_i0 + _fclamp(u, 0.0, 1.0)
                        src_j = src_j0 + _fclamp(v, 0.0, 1.0)
                if src_i == -1 and det_b != 0.0:
                    # noinspection PyTypeChecker
                    u = _fu(dst_x, dst_y, dst_p3x, dst_p3y, dst_p1x, dst_p1y) / det_b
                    # noinspection PyTypeChecker
                    v = _fv(dst_x, dst_y, dst_p3x, dst_p3y, dst_p2x, dst_p2y) / det_b
                    if u >= u_min and v >= v_min and u + v <= uv_max:
                        src_i = src_i1 - _fclamp(u, 0.0, 1.0)
                        src_j = src_j1 - _fclamp(v, 0.0, 1.0)
                if src_i != -1:
                    dst_src_ij_images[0, dst_j, dst_i] = src_i_min + src_i
                    dst_src_ij_images[1, dst_j, dst_i] = src_j_min + src_j


def _compute_var_image_xarray_numpy(
    src_var: xr.DataArray,
    dst_src_ij_images: np.ndarray,
    fill_value: Union[int, float, complex] = np.nan,
    interpolation: int = 0,
) -> np.ndarray:
    """Extract source pixels from xarray.DataArray source
    with numpy.ndarray data.
    """
    return _compute_var_image_numpy(
        src_var.values, dst_src_ij_images, fill_value, interpolation
    )


def _compute_var_image_xarray_dask(
    src_var: xr.DataArray,
    dst_src_ij_images: da.Array,
    fill_value: Union[int, float, complex] = np.nan,
    interpolation: int = 0,
) -> da.Array:
    """Extract source pixels from xarray.DataArray source
    with dask.array.Array data.
    """
    # If the source variable is not 3D, dummy variables are added to ensure
    # that `_compute_var_image_xarray_dask_block` always operates on 3D chunks
    # when processing each Dask chunk.
    if src_var.ndim == 1:
        src_var = src_var.expand_dims(dim={"dummy0": 1, "dummy1": 1})
    if src_var.ndim == 2:
        src_var = src_var.expand_dims(dim={"dummy": 1})
    # Retrieve the chunk size required for `da.map_blocks`, as the resulting array
    # will have a different shape.
    chunksize = src_var.shape[:-2] + dst_src_ij_images.chunksize[-2:]
    arr = da.map_blocks(
        _compute_var_image_xarray_dask_block,
        dst_src_ij_images,
        src_var,
        fill_value,
        interpolation,
        chunksize,
        dtype=src_var.dtype,
        chunks=chunksize,
    )
    arr = arr[..., : dst_src_ij_images.shape[-2], : dst_src_ij_images.shape[-1]]
    if arr.shape[0] == 1:
        arr = arr[0, :, :]
    if arr.shape[0] == 1:
        arr = arr[0, :]
    return arr


@nb.njit(nogil=True, cache=True)
def _compute_var_image_numpy(
    src_var: np.ndarray,
    dst_src_ij_images: np.ndarray,
    fill_value: Union[int, float, complex],
    interpolation: int,
) -> np.ndarray:
    """Extract source pixels from numpy.ndarray source
    with numba in parallel mode.
    """
    dst_width = dst_src_ij_images.shape[-1]
    dst_height = dst_src_ij_images.shape[-2]
    dst_shape = src_var.shape[:-2] + (dst_height, dst_width)
    dst_values = np.full(dst_shape, fill_value, dtype=src_var.dtype)
    src_bbox = (0, 0, src_var.shape[-2], src_var.shape[-1])
    _compute_var_image_numpy_parallel(
        src_var, dst_src_ij_images, dst_values, src_bbox, interpolation
    )
    return dst_values


def _compute_var_image_xarray_dask_block(
    dst_src_ij_images: np.ndarray,
    src_var_image: xr.DataArray,
    fill_value: Union[int, float, complex],
    interpolation: int,
    chunksize: tuple[int],
) -> np.ndarray:
    """Extract source pixels from np.ndarray source
    and return a block of a dask array.
    """
    dst_width = dst_src_ij_images.shape[-1]
    dst_height = dst_src_ij_images.shape[-2]
    dst_shape = src_var_image.shape[:-2] + (dst_height, dst_width)
    dst_out = np.full(chunksize, fill_value, dtype=src_var_image.dtype)
    if np.all(np.isnan(dst_src_ij_images[0])):
        return dst_out
    dst_values = np.full(dst_shape, fill_value, dtype=src_var_image.dtype)
    src_bbox = (
        int(np.nanmin(dst_src_ij_images[0])),
        int(np.nanmin(dst_src_ij_images[1])),
        min(int(np.nanmax(dst_src_ij_images[0])) + 2, src_var_image.shape[-1]),
        min(int(np.nanmax(dst_src_ij_images[1])) + 2, src_var_image.shape[-2]),
    )
    src_var_image = src_var_image[
        ..., src_bbox[1] : src_bbox[3], src_bbox[0] : src_bbox[2]
    ].values.astype(np.float64)
    _compute_var_image_numpy_sequential(
        src_var_image, dst_src_ij_images, dst_values, src_bbox, interpolation
    )
    dst_out[..., :dst_height, :dst_width] = dst_values
    return dst_out


@nb.njit(nogil=True, parallel=True, cache=True)
def _compute_var_image_numpy_parallel(
    src_var_image: np.ndarray,
    dst_src_ij_images: np.ndarray,
    dst_var_image: np.ndarray,
    src_bbox: tuple[int, int, int, int],
    interpolation: int,
):
    """Extract source pixels from np.ndarray source
    using numba parallel mode.
    """
    dst_height = dst_var_image.shape[-2]
    for dst_j in nb.prange(dst_height):
        _compute_var_image_for_dest_line(
            dst_j,
            src_var_image,
            dst_src_ij_images,
            dst_var_image,
            src_bbox,
            interpolation,
        )


# Extra dask version, because if we use parallel=True
# and nb.prange, we end up in infinite JIT compilation :(
@nb.njit(nogil=True, cache=True)
def _compute_var_image_numpy_sequential(
    src_var_image: np.ndarray,
    dst_src_ij_images: np.ndarray,
    dst_var_image: np.ndarray,
    src_bbox: tuple[int, int, int, int],
    interpolation: int,
):
    """Extract source pixels from np.ndarray source
    NOT using numba parallel mode.
    """
    dst_height = dst_var_image.shape[-2]
    for dst_j in range(dst_height):
        _compute_var_image_for_dest_line(
            dst_j,
            src_var_image,
            dst_src_ij_images,
            dst_var_image,
            src_bbox,
            interpolation,
        )


@nb.njit(nogil=True, cache=True)
def _compute_var_image_for_dest_line(
    dst_j: int,
    src_var_image: np.ndarray,
    dst_src_ij_images: np.ndarray,
    dst_var_image: np.ndarray,
    src_bbox: tuple[int, int, int, int],
    interpolation: int,
):
    """Extract source pixels from *src_values* np.ndarray
    and write into dst_values np.ndarray.
    """
    src_width = src_var_image.shape[-1]
    src_height = src_var_image.shape[-2]
    dst_width = dst_var_image.shape[-1]
    src_i_min = 0
    src_j_min = 0
    src_i_max = src_width - 1
    src_j_max = src_height - 1
    for dst_i in range(dst_width):
        src_i_f = dst_src_ij_images[0, dst_j, dst_i] - src_bbox[0]
        src_j_f = dst_src_ij_images[1, dst_j, dst_i] - src_bbox[1]
        if np.isnan(src_i_f) or np.isnan(src_j_f):
            continue
        # Note int() is 2x faster than math.floor() and
        # should yield the same results for only positive i,j.
        src_i0 = int(src_i_f)
        src_j0 = int(src_j_f)
        u = src_i_f - src_i0
        v = src_j_f - src_j0
        if interpolation == 0:
            # interpolation == "nearest"
            if u > 0.5:
                src_i0 = _iclamp(src_i0 + 1, src_i_min, src_i_max)
            if v > 0.5:
                src_j0 = _iclamp(src_j0 + 1, src_j_min, src_j_max)
            dst_var_value = src_var_image[..., src_j0, src_i0]
        elif interpolation == 1:
            # interpolation == "triangular"
            src_i1 = _iclamp(src_i0 + 1, src_i_min, src_i_max)
            src_j1 = _iclamp(src_j0 + 1, src_j_min, src_j_max)
            value_01 = src_var_image[..., src_j0, src_i1]
            value_10 = src_var_image[..., src_j1, src_i0]
            if u + v < 1.0:
                # Closest triangle
                value_00 = src_var_image[..., src_j0, src_i0]
                dst_var_value = (
                    value_00 + u * (value_01 - value_00) + v * (value_10 - value_00)
                )
            else:
                # Opposite triangle
                value_11 = src_var_image[..., src_j1, src_i1]
                dst_var_value = (
                    value_11
                    + (1.0 - u) * (value_10 - value_11)
                    + (1.0 - v) * (value_01 - value_11)
                )
        else:
            # interpolation == "bilinear"
            src_i1 = _iclamp(src_i0 + 1, src_i_min, src_i_max)
            src_j1 = _iclamp(src_j0 + 1, src_j_min, src_j_max)
            value_00 = src_var_image[..., src_j0, src_i0]
            value_01 = src_var_image[..., src_j0, src_i1]
            value_10 = src_var_image[..., src_j1, src_i0]
            value_11 = src_var_image[..., src_j1, src_i1]
            value_u0 = value_00 + u * (value_01 - value_00)
            value_u1 = value_10 + u * (value_11 - value_10)
            dst_var_value = value_u0 + v * (value_u1 - value_u0)
        dst_var_image[..., dst_j, dst_i] = dst_var_value


@nb.njit(
    "float64(float64, float64, float64, float64, float64, float64)",
    nogil=True,
    inline="always",
)
def _fdet(
    px0: float, py0: float, px1: float, py1: float, px2: float, py2: float
) -> float:
    return (px0 - px1) * (py0 - py2) - (px0 - px2) * (py0 - py1)


@nb.njit(
    "float64(float64, float64, float64, float64, float64, float64)",
    nogil=True,
    inline="always",
)
def _fu(px: float, py: float, px0: float, py0: float, px2: float, py2: float) -> float:
    return (px0 - px) * (py0 - py2) - (py0 - py) * (px0 - px2)


@nb.njit(
    "float64(float64, float64, float64, float64, float64, float64)",
    nogil=True,
    inline="always",
)
def _fv(px: float, py: float, px0: float, py0: float, px1: float, py1: float) -> float:
    return (py0 - py) * (px0 - px1) - (px0 - px) * (py0 - py1)


@nb.njit("float64(float64, float64, float64)", nogil=True, inline="always")
def _fclamp(x: float, x_min: float, x_max: float) -> float:
    return x_min if x < x_min else (x_max if x > x_max else x)


@nb.njit("int64(int64, int64, int64)", nogil=True, inline="always")
def _iclamp(x: int, x_min: int, x_max: int) -> int:
    return x_min if x < x_min else (x_max if x > x_max else x)


def _millis(seconds: float) -> int:
    return round(1000 * seconds)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonBooleanSchema,
    JsonNumberSchema,
    JsonStringSchema,
)

BOOLEAN_SCHEMA = JsonBooleanSchema()
NUMBER_SCHEMA = JsonNumberSchema()
URI_SCHEMA = JsonStringSchema(format="uri")
IDENTIFIER_SCHEMA = JsonStringSchema(min_length=1)
CHUNK_SIZE_SCHEMA = JsonStringSchema(min_length=2)  # TODO: use pattern
STRING_SCHEMA = JsonStringSchema()
PATH_SCHEMA = JsonStringSchema(min_length=1)

BOUNDING_BOX_ITEMS = [NUMBER_SCHEMA, NUMBER_SCHEMA, NUMBER_SCHEMA, NUMBER_SCHEMA]

BOUNDING_BOX_SCHEMA = JsonArraySchema(
    items=BOUNDING_BOX_ITEMS,
    description="Spatial bounding box given as [x-min, y-min, x-max, y-max].",
)

GEO_BOUNDING_BOX_SCHEMA = JsonArraySchema(
    items=BOUNDING_BOX_ITEMS,
    description="Spatial bounding box given as [x-min, y-min, x-max, y-max]"
    " using geographical coordinates.",
)

FILE_SYSTEM_SCHEMA = JsonStringSchema(enum=["memory", "obs", "local", "s3", "file"])



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.core.varexpr import VarExprContext, VarExprError

from ...server.api import ApiError
from .context import ExpressionsContext


# noinspection PyUnusedLocal
def get_expressions_capabilities(ctx: ExpressionsContext):
    return dict(
        namespace=dict(
            constants=VarExprContext.get_constants(),
            arrayFunctions=VarExprContext.get_array_functions(),
            otherFunctions=VarExprContext.get_other_functions(),
            arrayOperators=VarExprContext.get_array_operators(),
            otherOperators=VarExprContext.get_other_operators(),
        )
    )


def validate_expression(ctx: ExpressionsContext, ds_id: str, var_expr: str):
    dataset = ctx.datasets_ctx.get_dataset(ds_id)
    indexers = {
        str(dim_name): slice(0, min(2, dim_size))
        for dim_name, dim_size in dataset.sizes.items()
        if dim_size > 0
        and dim_name in dataset.coords
        and dataset.coords[dim_name].shape == (dim_size,)
    }
    var_expr_ctx = VarExprContext(dataset.isel(indexers))
    try:
        result = var_expr_ctx.evaluate(var_expr)
    except VarExprError as e:
        raise ApiError.BadRequest(f"{e}")
    try:
        return dict(result=result.mean().values.item())
    except BaseException as e:
        raise ApiError.BadRequest(f"{e}")



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import ABC, abstractmethod

import xarray as xr

from xcube.util.jsonschema import JsonObjectSchema

# Name of the (instance) method that processes a
# dataset and returns a new one.
# The one and only argument of the method is an xarray.Dataset.
# Then any number of processing parameters can be passed as
# keyword-arguments.
# See :meth:`DatasetProcessor#process_dataset`.
METHOD_NAME_DATASET_PROCESSOR = "process_dataset"

# Name of the (static/class) method that returns
# the JSON Object Schema for the processor parameters.
# The methods takes no arguments.
# See :meth:`DatasetProcessor#get_process_params_schema`.
METHOD_NAME_PARAMS_SCHEMA_GETTER = "get_process_params_schema"


class DatasetProcessor(ABC):
    """A generic dataset processor."""

    @abstractmethod
    def process_dataset(self, dataset: xr.Dataset, **process_params) -> xr.Dataset:
        """Process *dataset* into a new dataset using the
        given processing parameters *process_params*.

        The passed *process_params* must validate against
        the JSON schema returned by the :meth:`get_process_params_schema`
        class method.

        Args:
            dataset: The dataset to be processed.
            **process_params: The processing parameters.

        Returns:
            A new dataset.
        """

    @classmethod
    def get_process_params_schema(cls) -> JsonObjectSchema:
        """Get the JSON Schema for the processing parameters passed
        to :meth:`process_dataset`.

        The parameters are described by the JSON Object Schema's
        properties.

        The default implementation returns a JSON Schema
        that allows for any parameters of any type.

        Returns:
            a JSON Object Schema.
        """
        return JsonObjectSchema(additional_properties=True)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import JsonIntegerSchema, JsonObjectSchema

DEFAULT_MAX_VOXEL_COUNT = 256**3

VOLUMES_ACCESS_SCHEMA = JsonObjectSchema(
    properties=dict(
        MaxVoxelCount=JsonIntegerSchema(
            minimum=10**3,
            default=DEFAULT_MAX_VOXEL_COUNT,
        ),
    ),
    additional_properties=False,
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        VolumesAccess=VOLUMES_ACCESS_SCHEMA,
    )
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from collections.abc import Sequence
from typing import Any, Dict, Union

import numpy as np
import xarray as xr

from xcube.core.schema import CubeSchema
from xcube.core.select import select_variables_subset
from xcube.core.verify import assert_cube


def resample_in_time(
    dataset: xr.Dataset,
    frequency: str,
    method: Union[str, Sequence[str]],
    offset=None,
    tolerance=None,
    interp_kind=None,
    time_chunk_size=None,
    var_names: Sequence[str] = None,
    metadata: dict[str, Any] = None,
    cube_asserted: bool = False,
) -> xr.Dataset:
    """Resample a dataset in the time dimension.

    The argument *method* may be one or a sequence of
    ``'all'``, ``'any'``,
    ``'argmax'``, ``'argmin'``, ``'count'``,
    ``'first'``, ``'last'``,
    ``'max'``, ``'min'``, ``'mean'``, ``'median'``,
    ``'percentile_<p>'``,
    ``'std'``, ``'sum'``, ``'var'``.

    In value ``'percentile_<p>'`` is a placeholder,
    where ``'<p>'`` must be replaced by an integer percentage
    value, e.g. ``'percentile_90'`` is the 90%-percentile.

    *Important note:* As of xarray 0.14 and dask 2.8, the
    methods ``'median'`` and ``'percentile_<p>'` cannot be
    used if the variables in *cube* comprise chunked dask arrays.
    In this case, use the ``compute()`` or ``load()`` method
    to convert dask arrays into numpy arrays.

    Args:
        dataset: The xcube dataset.
        frequency: Temporal aggregation frequency. Use format
            "<count><offset>" where <offset> is one of 'H', 'D', 'W',
            'M', 'Q', 'Y'.
        method: Resampling method or sequence of resampling methods.
        offset: Offset used to adjust the resampled time labels. Uses
            same syntax as *frequency*.
        time_chunk_size: If not None, the chunk size to be used for the
            "time" dimension.
        var_names: Variable names to include.
        tolerance: Time tolerance for selective upsampling methods.
            Defaults to *frequency*.
        interp_kind: Kind of interpolation if *method* is
            'interpolation'.
        metadata: Output metadata.
        cube_asserted: If False, *cube* will be verified, otherwise it
            is expected to be a valid cube.

    Returns:
        A new xcube dataset resampled in time.
    """
    if not cube_asserted:
        assert_cube(dataset)

    if frequency == "all":
        time_gap = np.array(dataset.time[-1]) - np.array(dataset.time[0])
        days = int((np.timedelta64(time_gap, "D") / np.timedelta64(1, "D")) + 1)
        frequency = f"{days}D"

    if var_names:
        dataset = select_variables_subset(dataset, var_names)

    resampler = dataset.resample(
        skipna=True, closed="left", label="left", time=frequency, offset=offset
    )

    if isinstance(method, str):
        methods = [method]
    else:
        methods = list(method)

    percentile_prefix = "percentile_"

    resampled_cubes = []
    for method in methods:
        method_args = []
        method_postfix = method
        if method.startswith(percentile_prefix):
            p = int(method[len(percentile_prefix) :])
            q = p / 100.0
            method_args = [q]
            method_postfix = f"p{p}"
            method = "quantile"
        resampling_method = getattr(resampler, method)
        method_kwargs = get_method_kwargs(method, frequency, interp_kind, tolerance)
        resampled_cube = resampling_method(*method_args, **method_kwargs)
        resampled_cube = resampled_cube.rename(
            {
                var_name: f"{var_name}_{method_postfix}"
                for var_name in resampled_cube.data_vars
            }
        )
        resampled_cubes.append(resampled_cube)

    if len(resampled_cubes) == 1:
        resampled_cube = resampled_cubes[0]
    else:
        resampled_cube = xr.merge(resampled_cubes)

    # TODO: add time_bnds to resampled_ds
    time_coverage_start = "%s" % dataset.time[0]
    time_coverage_end = "%s" % dataset.time[-1]

    resampled_cube.attrs.update(metadata or {})
    # TODO: add other time_coverage_ attributes
    resampled_cube.attrs.update(
        time_coverage_start=time_coverage_start, time_coverage_end=time_coverage_end
    )

    schema = CubeSchema.new(dataset)
    chunk_sizes = {schema.dims[i]: schema.chunks[i] for i in range(schema.ndim)}

    if isinstance(time_chunk_size, int) and time_chunk_size >= 0:
        chunk_sizes["time"] = time_chunk_size

    return resampled_cube.chunk(chunk_sizes)


def get_method_kwargs(method, frequency, interp_kind, tolerance):
    if method == "interpolate":
        kwargs = {"kind": interp_kind or "linear"}
    elif method in {"nearest", "bfill", "ffill", "pad"}:
        kwargs = {"tolerance": tolerance or frequency}
    elif method in {
        "first",
        "last",
        "sum",
        "min",
        "max",
        "mean",
        "median",
        "std",
        "var",
    }:
        kwargs = {"dim": "time", "keep_attrs": True, "skipna": True}
    elif method == "prod":
        kwargs = {"dim": "time", "skipna": True}
    elif method == "count":
        kwargs = {"dim": "time", "keep_attrs": True}
    else:
        kwargs = {}
    return kwargs



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiHandler

from ..datasets.routes import PATH_PARAM_DATASET_ID
from .api import api
from .context import ExpressionsContext
from .controllers import get_expressions_capabilities, validate_expression

PATH_PARAM_VAR_EXPR = {
    "name": "varExpr",
    "in": "path",
    "description": "Variable expression",
    "schema": {"type": "string"},
}


@api.route("/expressions/capabilities")
class ExpressionsNamespaceHandler(ApiHandler[ExpressionsContext]):
    @api.operation(
        operation_id="getExpressionCapabilities",
        summary=(
            "Gets the server capabilities for expressions that define user variables"
        ),
    )
    def get(self):
        self.response.finish(get_expressions_capabilities(self.ctx))


# noinspection PyPep8Naming
@api.route("/expressions/validate/{datasetId}/{varExpr}")
class ExpressionsNamespaceHandler(ApiHandler[ExpressionsContext]):
    @api.operation(
        operation_id="validateExpression",
        summary="Validate the given variable expression",
        parameters=[PATH_PARAM_DATASET_ID, PATH_PARAM_VAR_EXPR],
    )
    def get(self, datasetId: str, varExpr: str):
        self.response.finish(validate_expression(self.ctx, datasetId, varExpr))



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import json
from typing import Any, Dict, List, Union

import numpy as np

JsonArray = list["JsonValue"]
JsonObject = dict[str, "JsonValue"]
JsonValue = Union[None, bool, int, float, str, JsonArray, JsonObject]


class NumpyJSONEncoder(json.JSONEncoder):
    """A JSON encoder that converts numpy-like
    scalars into corresponding serializable Python objects.
    """

    def default(self, obj: Any) -> JsonValue:
        converted_obj = _convert_default(obj)
        if converted_obj is not obj:
            return converted_obj
        return json.JSONEncoder.default(self, obj)


_PRIMITIVE_JSON_TYPES = {type(None), bool, int, float, str}


def to_json_value(obj: Any) -> JsonValue:
    """Convert *obj* into a JSON-serializable object.

    Args:
        obj: A Python object.

    Returns: A JSON-serializable version of *obj*, or *obj*
        if *obj* is already JSON-serializable.

    Raises:
        TypeError: If *obj* cannot be made JSON-serializable
    """
    if obj is None:
        return None

    converted_obj = _convert_default(obj)
    if converted_obj is not obj:
        return converted_obj

    obj_type = type(obj)

    if obj_type in _PRIMITIVE_JSON_TYPES:
        return obj

    for t in _PRIMITIVE_JSON_TYPES:
        if isinstance(obj, t):
            return t(obj)

    if obj_type is dict:
        converted_obj = {_key(k): to_json_value(v) for k, v in obj.items()}
        if any(converted_obj[k] is not obj[k] for k in obj.keys()):
            return converted_obj
        else:
            return obj

    if obj_type is list:
        converted_obj = [to_json_value(item) for item in obj]
        if any(o1 is not o2 for o1, o2 in zip(converted_obj, obj)):
            return converted_obj
        else:
            return obj

    try:
        return {_key(k): to_json_value(v) for k, v in obj.items()}
    except AttributeError:
        try:
            return [to_json_value(item) for item in obj]
        except TypeError:
            # Same as json.JSONEncoder.default(self, obj)
            raise TypeError(
                f"Object of type {obj.__class__.__name__} is not JSON serializable"
            )


def _key(key: Any) -> str:
    if not isinstance(key, str):
        raise TypeError(
            f"Property names of JSON objects must be strings,"
            f" but got {key.__class__.__name__}"
        )
    return key


def _convert_default(obj: Any) -> Any:
    if hasattr(obj, "dtype") and hasattr(obj, "ndim"):
        if obj.ndim == 0:
            if np.issubdtype(obj.dtype, bool):
                return bool(obj)
            elif np.issubdtype(obj.dtype, np.integer):
                return int(obj)
            elif np.issubdtype(obj.dtype, np.floating):
                return float(obj)
            elif np.issubdtype(obj.dtype, np.datetime64):
                # np.datetime_as_string returns a np._str object;
                # the str() then returns a str object.
                return str(np.datetime_as_string(obj, timezone="UTC", unit="s"))
            elif np.issubdtype(obj.dtype, np.str):
                return str(obj)
        else:
            return [_convert_default(item) for item in obj]
    # We may handle other non-JSON-serializable datatypes here
    return obj



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.util.jsonschema import JsonArraySchema, JsonObjectSchema
from xcube.webapi.common.schemas import STRING_SCHEMA

CONFIGURATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Path=STRING_SCHEMA,
    ),
    additional_properties=False,
)

PERSISTENCE_SCHEMA = JsonObjectSchema(
    properties=dict(
        Path=STRING_SCHEMA, StorageOptions=JsonObjectSchema(additional_properties=True)
    ),
    required=["Path"],
    additional_properties=False,
)

EXTENSIONS_SCHEMA = JsonArraySchema(
    items=STRING_SCHEMA,
    min_items=1,
)

AUGMENTATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Path=STRING_SCHEMA,
        Extensions=EXTENSIONS_SCHEMA,
    ),
    required=["Extensions"],
    additional_properties=False,
)

VIEWER_SCHEMA = JsonObjectSchema(
    properties=dict(
        Configuration=CONFIGURATION_SCHEMA,
        Augmentation=AUGMENTATION_SCHEMA,
        Persistence=PERSISTENCE_SCHEMA,
    ),
    additional_properties=False,
)

CONFIG_SCHEMA = JsonObjectSchema(
    properties=dict(
        Viewer=VIEWER_SCHEMA,
    )
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import cProfile
import glob
import io
import os
import pstats
import sys
import time
import traceback
import warnings
from collections.abc import Sequence
from typing import Any, Callable, Dict, Tuple

import xarray as xr

from xcube.core.dsio import DatasetIO, find_dataset_io, guess_dataset_format, rimraf
from xcube.core.evaluate import evaluate_dataset
from xcube.core.gen.defaults import (
    DEFAULT_OUTPUT_PATH,
    DEFAULT_OUTPUT_RESAMPLING,
    DEFAULT_OUTPUT_SIZE,
)
from xcube.core.gen.iproc import InputProcessor, find_input_processor_class
from xcube.core.gridmapping import CRS_WGS84, GridMapping
from xcube.core.optimize import optimize_dataset
from xcube.core.select import select_spatial_subset, select_variables_subset
from xcube.core.timecoord import add_time_coords, from_time_in_days_since_1970
from xcube.core.timeslice import find_time_slice
from xcube.core.update import (
    update_dataset_attrs,
    update_dataset_temporal_attrs,
    update_dataset_var_attrs,
)
from xcube.util.config import NameAnyDict, NameDictPairList, to_resolved_name_dict_pairs


def gen_cube(
    input_paths: Sequence[str] = None,
    input_processor_name: str = None,
    input_processor_params: dict = None,
    input_reader_name: str = None,
    input_reader_params: dict[str, Any] = None,
    output_region: tuple[float, float, float, float] = None,
    output_size: tuple[int, int] = DEFAULT_OUTPUT_SIZE,
    output_resampling: str = DEFAULT_OUTPUT_RESAMPLING,
    output_path: str = DEFAULT_OUTPUT_PATH,
    output_writer_name: str = None,
    output_writer_params: dict[str, Any] = None,
    output_metadata: NameAnyDict = None,
    output_variables: NameDictPairList = None,
    processed_variables: NameDictPairList = None,
    profile_mode: bool = False,
    no_sort_mode: bool = False,
    append_mode: bool = None,
    dry_run: bool = False,
    monitor: Callable[..., None] = None,
) -> bool:
    """Generate a xcube dataset from one or more input files.

    Args:
        no_sort_mode
        input_paths: The input paths.
        input_processor_name: Name of a registered input processor
            (xcube.core.gen.inputprocessor.InputProcessor) to be used to
            transform the inputs.
        input_processor_params: Parameters to be passed to the input
            processor.
        input_reader_name: Name of a registered input reader
            (xcube.core.util.dsio.DatasetIO).
        input_reader_params: Parameters passed to the input reader.
        output_region: Output region as tuple of floats: (lon_min,
            lat_min, lon_max, lat_max).
        output_size: The spatial dimensions of the output as tuple of
            ints: (width, height).
        output_resampling: The resampling method for the output.
        output_path: The output directory.
        output_writer_name: Name of an output writer
            (xcube.core.util.dsio.DatasetIO) used to write the cube.
        output_writer_params: Parameters passed to the output writer.
        output_metadata: Extra metadata passed to output cube.
        output_variables: Output variables.
        processed_variables: Processed variables computed on-the-fly.
        profile_mode: Whether profiling should be enabled.
        append_mode: Deprecated. The function will always either insert,
            replace, or append new time slices.
        dry_run: Doesn't write any data. For testing.
        monitor: A progress monitor.

    Returns:
        True for success.
    """

    if append_mode is not None:
        warnings.warn(
            "append_mode in gen_cube() is deprecated, "
            "time slices will now always be inserted, replaced, or appended."
        )

    if input_processor_name is None:
        input_processor_name = "default"
    elif input_processor_name == "":
        raise ValueError("input_processor_name must not be empty")

    input_processor_class = find_input_processor_class(input_processor_name)
    if not input_processor_class:
        raise ValueError(f"Unknown input_processor_name {input_processor_name!r}")

    if not issubclass(input_processor_class, InputProcessor):
        raise ValueError(
            f"Invalid input_processor_name {input_processor_name!r}: "
            f"must name a sub-class of {InputProcessor.__qualname__}"
        )

    try:
        input_processor = input_processor_class(**(input_processor_params or {}))
    except (ValueError, TypeError) as e:
        raise ValueError(
            f"Invalid input_processor_name or input_processor_params: {e}"
        ) from e

    input_reader = find_dataset_io(input_reader_name or input_processor.input_reader)
    if not input_reader:
        raise ValueError(f"Unknown input_reader_name {input_reader_name!r}")

    if not output_path:
        raise ValueError("Missing output_path")

    output_writer_name = output_writer_name or guess_dataset_format(output_path)
    if not output_writer_name:
        raise ValueError(f"Failed to guess output_writer_name from path {output_path}")
    output_writer = find_dataset_io(output_writer_name, modes={"w", "a"})
    if not output_writer:
        raise ValueError(f"Unknown output_writer_name {output_writer_name!r}")

    if monitor is None:
        # noinspection PyUnusedLocal
        def monitor(*args):
            pass

    input_paths = [
        input_file for f in input_paths for input_file in glob.glob(f, recursive=True)
    ]

    effective_input_reader_params = dict(input_processor.input_reader_params or {})
    effective_input_reader_params.update(input_reader_params or {})

    if not no_sort_mode and len(input_paths) > 1:
        input_paths = _get_sorted_input_paths(
            input_processor, input_reader, effective_input_reader_params, input_paths
        )

    if not dry_run:
        output_dir = os.path.abspath(os.path.dirname(output_path))
        os.makedirs(output_dir, exist_ok=True)

    effective_output_writer_params = output_writer_params or {}

    status = False

    ds_count = len(input_paths)
    ds_count_ok = 0
    ds_index = 0
    for input_file in input_paths:
        monitor(f"processing dataset {ds_index + 1} of {ds_count}: {input_file!r}...")
        # noinspection PyTypeChecker
        status = _process_input(
            input_processor,
            input_reader,
            effective_input_reader_params,
            output_writer,
            effective_output_writer_params,
            input_file,
            output_size,
            output_region,
            output_resampling,
            output_path,
            output_metadata,
            output_variables,
            processed_variables,
            profile_mode,
            dry_run,
            monitor,
        )
        ds_index += 1
        if status:
            ds_count_ok += 1

    monitor(
        f"{ds_count_ok} of {ds_count} datasets processed successfully, "
        f"{ds_count - ds_count_ok} were dropped due to errors"
    )

    return status


def _process_input(
    input_processor: InputProcessor,
    input_reader: DatasetIO,
    input_reader_params: dict[str, Any],
    output_writer: DatasetIO,
    output_writer_params: dict[str, Any],
    input_file: str,
    output_size: tuple[int, int],
    output_region: tuple[float, float, float, float],
    output_resampling: str,
    output_path: str,
    output_metadata: NameAnyDict = None,
    output_variables: NameDictPairList = None,
    processed_variables: NameDictPairList = None,
    profile_mode: bool = False,
    dry_run: bool = False,
    monitor: Callable[..., None] = None,
) -> bool:
    monitor("reading input slice...")
    # noinspection PyBroadException
    try:
        input_dataset = input_reader.read(input_file, **input_reader_params)
        monitor(f"Dataset read:\n{input_dataset}")
    except Exception as e:
        monitor(f"Error: cannot read input: {e}: skipping...")
        traceback.print_exc(file=sys.stderr)
        return False

    time_range = input_processor.get_time_range(input_dataset)
    if time_range[0] > time_range[1]:
        monitor("Error: start time is greater than end time: skipping...")
        return False

    if output_variables:
        output_variables = to_resolved_name_dict_pairs(
            output_variables, input_dataset, keep=True
        )
    else:
        output_variables = [(var_name, None) for var_name in input_dataset.data_vars]

    time_index, update_mode = find_time_slice(
        output_path, from_time_in_days_since_1970((time_range[0] + time_range[1]) / 2)
    )

    width, height = output_size
    x_min, y_min, x_max, y_max = output_region
    xy_res = max((x_max - x_min) / width, (y_max - y_min) / height)
    tile_size = _get_tile_size(output_writer_params)

    output_geom = GridMapping.regular(
        size=output_size,
        xy_min=(x_min, y_min),
        xy_res=xy_res,
        crs=CRS_WGS84,
        tile_size=tile_size,
    )

    steps = []

    # noinspection PyShadowingNames
    def step1(input_slice):
        return input_processor.pre_process(input_slice)

    steps.append((step1, "pre-processing input slice"))

    grid_mapping = None

    # noinspection PyShadowingNames
    def step1a(input_slice):
        nonlocal grid_mapping
        grid_mapping = GridMapping.from_dataset(input_slice)
        subset = select_spatial_subset(
            input_slice,
            xy_bbox=output_geom.xy_bbox,
            xy_border=output_geom.x_res,
            ij_border=1,
            grid_mapping=grid_mapping,
        )
        if subset is None:
            monitor("no spatial overlap with input")
        elif subset is not input_slice:
            grid_mapping = GridMapping.from_dataset(subset)
        return subset

    steps.append((step1a, "spatial subsetting"))

    # noinspection PyShadowingNames
    def step2(input_slice):
        return evaluate_dataset(input_slice, processed_variables=processed_variables)

    steps.append((step2, "computing input slice variables"))

    # noinspection PyShadowingNames
    def step3(input_slice):
        extra_vars = input_processor.get_extra_vars(input_slice)
        selected_variables = {var_name for var_name, _ in output_variables}
        selected_variables.update(extra_vars or set())
        return select_variables_subset(input_slice, selected_variables)

    steps.append((step3, "selecting input slice variables"))

    # noinspection PyShadowingNames
    def step4(input_slice):
        # noinspection PyTypeChecker
        return input_processor.process(
            input_slice,
            geo_coding=grid_mapping,
            output_geom=output_geom,
            output_resampling=output_resampling,
            include_non_spatial_vars=False,
        )

    steps.append((step4, "transforming input slice"))

    if time_range is not None:

        def step5(input_slice):
            return add_time_coords(input_slice, time_range)

        steps.append((step5, "adding time coordinates to input slice"))

    def step6(input_slice):
        return update_dataset_var_attrs(input_slice, output_variables)

    steps.append((step6, "updating variable attributes of input slice"))

    def step7(input_slice):
        return input_processor.post_process(input_slice)

    steps.append((step7, "post-processing input slice"))

    if update_mode == "create":

        def step8(input_slice):
            if not dry_run:
                rimraf(output_path)
                output_writer.write(input_slice, output_path, **output_writer_params)
                _update_cube(
                    output_writer,
                    output_path,
                    global_attrs=output_metadata,
                    temporal_only=False,
                )
            return input_slice

        steps.append((step8, f"creating input slice in {output_path}"))

    elif update_mode == "append":

        def step8(input_slice):
            if not dry_run:
                output_writer.append(input_slice, output_path, **output_writer_params)
                _update_cube(output_writer, output_path, temporal_only=True)
            return input_slice

        steps.append((step8, f"appending input slice to {output_path}"))

    elif update_mode == "insert":

        def step8(input_slice):
            if not dry_run:
                output_writer.insert(input_slice, time_index, output_path)
                _update_cube(output_writer, output_path, temporal_only=True)
            return input_slice

        steps.append(
            (step8, f"inserting input slice before index {time_index} in {output_path}")
        )

    elif update_mode == "replace":

        def step8(input_slice):
            if not dry_run:
                output_writer.replace(input_slice, time_index, output_path)
                _update_cube(output_writer, output_path, temporal_only=True)
            return input_slice

        steps.append(
            (step8, f"replacing input slice at index {time_index} in {output_path}")
        )

    if profile_mode:
        pr = cProfile.Profile()
        pr.enable()

    status = True
    try:
        num_steps = len(steps)
        dataset = input_dataset
        total_t1 = time.perf_counter()
        for step_index in range(num_steps):
            transform, label = steps[step_index]
            step_t1 = time.perf_counter()
            monitor(f"step {step_index + 1} of {num_steps}: {label}...")
            dataset = transform(dataset)
            step_t2 = time.perf_counter()
            if dataset is None:
                monitor(
                    f"  {label} terminated after {step_t2 - step_t1} seconds, skipping input slice"
                )
                status = False
                break
            monitor(f"  {label} completed in {step_t2 - step_t1} seconds")
        total_t2 = time.perf_counter()
        monitor(f"{num_steps} steps took {total_t2 - total_t1} seconds to complete")
    except RuntimeError as e:
        monitor(
            f"Error: something went wrong during processing, skipping input slice: {e}"
        )
        traceback.print_exc(file=sys.stderr)
        status = False
    finally:
        input_dataset.close()

    if profile_mode:
        # noinspection PyUnboundLocalVariable
        pr.disable()
        s = io.StringIO()
        ps = pstats.Stats(pr, stream=s).sort_stats("cumtime")
        ps.print_stats()
        monitor(s.getvalue())

    return status


def _get_tile_size(output_writer_params):
    tile_size = None
    if "chunksizes" in output_writer_params:
        chunksizes = output_writer_params["chunksizes"]
        if "lon" in chunksizes or "lat" in chunksizes:
            tile_size = chunksizes.get("lon", 512), chunksizes.get("lat", 512)
        elif "x" in chunksizes or "y" in chunksizes:
            tile_size = chunksizes.get("x", 512), chunksizes.get("y", 512)
    return tile_size


def _update_cube(
    output_writer: DatasetIO,
    output_path: str,
    global_attrs: dict = None,
    temporal_only: bool = False,
):
    cube = output_writer.read(output_path)
    if temporal_only:
        cube = update_dataset_temporal_attrs(cube, update_existing=True, in_place=True)
    else:
        cube = update_dataset_attrs(cube, update_existing=True, in_place=True)
    cube_attrs = dict(cube.attrs)
    cube.close()

    if global_attrs:
        cube_attrs.update(global_attrs)

    output_writer.update(output_path, global_attrs=cube_attrs)


def _get_sorted_input_paths(
    input_processor,
    input_reader: DatasetIO,
    input_reader_params: dict[str, Any],
    input_paths: Sequence[str],
):
    input_path_list = []
    time_list = []
    for input_file in input_paths:
        #        with xr.open_dataset(input_file) as dataset:
        with input_reader.read(input_file, **input_reader_params) as dataset:
            t1, t2 = input_processor.get_time_range(dataset)
            time_list.append((t1 + t2) / 2)
            input_path_list.append(input_file)
            tuple_seq = zip(time_list, input_path_list)
    input_paths = [e[1] for e in sorted(tuple_seq, key=lambda e: e[0])]
    return input_paths



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


import fnmatch
import itertools
import os
import os.path
import warnings
from collections.abc import Collection, Mapping
from functools import cached_property
from typing import Any, Callable, Optional, Union

import numpy as np
import pandas as pd
import pyproj
import xarray as xr

from xcube.constants import LOG
from xcube.core.mldataset import (
    BaseMultiLevelDataset,
    IdentityMultiLevelDataset,
    MultiLevelDataset,
    augment_ml_dataset,
    open_ml_dataset_from_python_code,
)
from xcube.core.store import (
    DATASET_TYPE,
    MULTI_LEVEL_DATASET_TYPE,
    DatasetDescriptor,
    DataStoreConfig,
    DataStorePool,
)
from xcube.core.tile import (
    DEFAULT_CMAP_NAME,
    DEFAULT_CMAP_NORM,
    DEFAULT_VALUE_RANGE,
    get_var_cmap_params,
    get_var_valid_range,
)
from xcube.core.varexpr import split_var_assignment
from xcube.server.api import ApiError, Context, ServerConfig
from xcube.server.config import is_absolute_path
from xcube.util.assertions import assert_given, assert_instance
from xcube.util.cache import parse_mem_size
from xcube.util.cmaps import (
    ColormapRegistry,
    create_colormap_from_config,
    load_custom_colormap,
)
from xcube.webapi.common.context import ResourcesContext
from xcube.webapi.places import PlacesContext

COMPUTE_DATASET = "compute_dataset"
COMPUTE_VARIABLES = "compute_variables"

# We use tilde, because it is not a reserved URI characters
STORE_DS_ID_SEPARATOR = "~"
DATA_STORE_IDS_WARNING = ["stac"]
FS_TYPE_TO_PROTOCOL = {
    "local": "file",
    "obs": "s3",
    "file": "file",
    "s3": "s3",
    "memory": "memory",
}
NON_MEMORY_FILE_SYSTEMS = ["local", "obs", "file", "s3"]

ALL_PLACES = "all"

MultiLevelDatasetOpener = Callable[["DatasetsContext", ServerConfig], MultiLevelDataset]

DatasetConfig = Mapping[str, Any]


class DatasetsContext(ResourcesContext):
    def __init__(self, server_ctx: Context, ml_dataset_openers=None):
        super().__init__(server_ctx)
        # noinspection PyTypeChecker
        self._places_ctx: PlacesContext = server_ctx.get_api_ctx("places")
        assert isinstance(self._places_ctx, PlacesContext)
        self._ml_dataset_openers = ml_dataset_openers
        # cache for all dataset configs
        # contains tuples of form (MultiLevelDataset, dataset_config)
        self._dataset_cache = dict()
        (
            self._data_store_pool,
            self._dataset_configs,
            self.entrypoint_dataset_id,
            self._dataset_groups_config,
        ) = self._process_dataset_configs(self.config, self.base_dir)
        self._cm_styles, self._colormap_registry = self._get_cm_styles()

    def on_dispose(self):
        with self.rlock:
            # Close all datasets
            for ml_dataset, _ in self._dataset_cache.values():
                # noinspection PyBroadException
                try:
                    ml_dataset.close()
                except Exception:
                    pass
            # Clear all caches
            if self._dataset_cache:
                self._dataset_cache.clear()
            if self._data_store_pool:
                self._data_store_pool.remove_all_store_configs()
            self._dataset_configs = None

    @property
    def places_ctx(self):
        return self._places_ctx

    @property
    def dataset_cache(self) -> dict[str, tuple[MultiLevelDataset, DatasetConfig]]:
        return self._dataset_cache

    @cached_property
    def access_control(self) -> dict[str, Any]:
        return self.config.get("AccessControl", {})

    @cached_property
    def required_scopes(self) -> list[str]:
        return self.access_control.get("RequiredScopes", [])

    @property
    def colormap_registry(self) -> ColormapRegistry:
        return self._colormap_registry

    def get_required_dataset_scopes(self, dataset_config: DatasetConfig) -> set[str]:
        return self._get_required_scopes(
            dataset_config, "read:dataset", "Dataset", dataset_config["Identifier"]
        )

    def get_required_variable_scopes(
        self, dataset_config: DatasetConfig, var_name: str
    ) -> set[str]:
        return self._get_required_scopes(
            dataset_config, "read:variable", "Variable", var_name
        )

    def _get_required_scopes(
        self,
        dataset_config: DatasetConfig,
        base_scope: str,
        value_name: str,
        value: str,
    ) -> set[str]:
        required_global_scopes = set(self.required_scopes)
        required_dataset_scopes = set(
            dataset_config.get("AccessControl", {}).get("RequiredScopes", [])
        )
        if not required_global_scopes and not required_dataset_scopes:
            return required_global_scopes
        required_dataset_scopes = required_global_scopes.union(required_dataset_scopes)
        base_scope_prefix = base_scope + ":"
        required_dataset_scopes = {
            scope
            for scope in required_dataset_scopes
            if scope == base_scope or scope.startswith(base_scope_prefix)
        }
        pattern_scope = base_scope_prefix + "{" + value_name + "}"
        if pattern_scope in required_dataset_scopes:
            # Replace "{base_scope}:{value_name}" by "{base_scope}:{value}"
            required_dataset_scopes.remove(pattern_scope)
            required_dataset_scopes.add(base_scope_prefix + value)
        return required_dataset_scopes

    def get_ml_dataset(self, ds_id: str) -> MultiLevelDataset:
        ml_dataset, _ = self._get_dataset_entry(ds_id)
        return ml_dataset

    def set_ml_dataset(self, ml_dataset: MultiLevelDataset):
        self._set_dataset_entry(
            (ml_dataset, _new_dataset_config(Identifier=ml_dataset.ds_id, Hidden=True))
        )

    def add_dataset(
        self,
        dataset: Union[xr.Dataset, MultiLevelDataset],
        ds_id: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        style: Optional[str] = None,
        color_mappings: dict[str, dict[str, Any]] = None,
    ) -> str:
        """Add a dataset to the configuration"""
        assert_instance(dataset, (xr.Dataset, MultiLevelDataset), "dataset")
        if isinstance(dataset, xr.Dataset):
            ml_dataset = BaseMultiLevelDataset(dataset, ds_id=ds_id)
        else:
            ml_dataset = dataset
            if not ds_id:
                _ds_id = ml_dataset.ds_id
                # TODO: Fixes https://github.com/xcube-dev/xcube/issues/1007, but
                #   this does not solve the likely root cause
                #   in xcube.core.mldataset.fs.FsMultiLevelDataset, see ctor.
                if "/" in _ds_id:
                    ds_id = _ds_id.replace("/", "-")
            if ds_id:
                ml_dataset = IdentityMultiLevelDataset(dataset, ds_id=ds_id)
            dataset = ml_dataset.base_dataset
        ds_id = ml_dataset.ds_id
        dataset_configs = self.get_dataset_configs()
        for index, dataset_config in enumerate(dataset_configs):
            if dataset_config["Identifier"] == ds_id:
                del dataset_configs[index]
                break
        dataset_config = _new_dataset_config(
            Identifier=ds_id, Title=title, Description=description
        )
        if style is not None:
            dataset_config.update(dict(Style=style))
        if color_mappings is not None:
            style = style or ds_id
            dataset_config.update(dict(Style=style))
            self._cm_styles[style] = color_mappings
        self._dataset_cache[ds_id] = ml_dataset, dataset_config
        self._dataset_configs.append(dataset_config)
        return ds_id

    def remove_dataset(self, ds_id: str):
        assert_instance(ds_id, str, "ds_id")
        assert_given(ds_id, "ds_id")
        if ds_id in self._dataset_cache:
            del self._dataset_cache[ds_id]
        self._dataset_configs = [
            dc for dc in self._dataset_configs if dc["Identifier"] != ds_id
        ]

    def add_ml_dataset(
        self,
        ml_dataset: MultiLevelDataset,
        ds_id: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ):
        if ds_id:
            ml_dataset.ds_id = ds_id
        else:
            ds_id = ml_dataset.ds_id
        self._set_dataset_entry(
            (
                ml_dataset,
                (
                    dict(
                        Identifier=ds_id,
                        Title=title,
                        Description=description,
                    )
                ),
            )
        )

    def get_dataset(
        self, ds_id: str, expected_var_names: Collection[str] = None
    ) -> xr.Dataset:
        ml_dataset, _ = self._get_dataset_entry(ds_id)
        dataset = ml_dataset.base_dataset
        if expected_var_names:
            for var_name in expected_var_names:
                if var_name not in dataset:
                    raise ApiError.NotFound(
                        f'Variable "{var_name}" not found in dataset "{ds_id}"'
                    )
        return dataset

    def get_time_series_dataset(self, ds_id: str, var_name: str = None) -> xr.Dataset:
        dataset_config = self.get_dataset_config(ds_id)
        ts_ds_name = dataset_config.get("TimeSeriesDataset", ds_id)
        try:
            # Try to get more efficient, time-chunked dataset
            return self.get_dataset(
                ts_ds_name, expected_var_names=[var_name] if var_name else None
            )
        except ApiError.NotFound:
            # This happens, if the dataset pointed to by 'TimeSeriesDataset'
            # does not contain the variable given by var_name.
            return self.get_dataset(
                ds_id, expected_var_names=[var_name] if var_name else None
            )

    def get_variable_for_z(
        self, ds_id: str, var_name: str, z_index: int
    ) -> xr.DataArray:
        ml_dataset = self.get_ml_dataset(ds_id)
        index = ml_dataset.num_levels - 1 - z_index
        if index < 0 or index >= ml_dataset.num_levels:
            raise ApiError.NotFound(
                f'Variable "{var_name}" has no z-index {z_index} in dataset "{ds_id}"'
            )
        dataset = ml_dataset.get_dataset(index)
        if var_name not in dataset:
            raise ApiError.NotFound(
                f'Variable "{var_name}" not found in dataset "{ds_id}"'
            )
        return dataset[var_name]

    @classmethod
    def _maybe_assign_store_instance_ids(
        cls,
        dataset_configs: list[dict[str, Any]],
        data_store_pool: DataStorePool,
        base_dir: str,
    ) -> None:
        assignable_dataset_configs = [
            dc
            for dc in dataset_configs
            if "StoreInstanceId" not in dc
            and dc.get("FileSystem", "file") in NON_MEMORY_FILE_SYSTEMS
        ]
        # split into sub-lists according to file system and
        # non-root store params
        config_lists = []
        for config in assignable_dataset_configs:
            store_params = cls._get_other_store_params_than_root(config)
            file_system = config.get("FileSystem", "file")
            appended = False
            for config_list in config_lists:
                if config_list[0] == file_system and config_list[1] == store_params:
                    config_list[2].append(config)
                    appended = True
                    break
            if not appended:
                config_lists.append((file_system, store_params, [config]))

        for file_system, store_params, config_list in config_lists:
            # Retrieve paths per configuration
            paths = [dc["Path"] for dc in config_list]
            # Determine common prefixes of paths (and call them roots)
            roots = _get_roots(paths)
            for root in roots:
                fs_protocol = FS_TYPE_TO_PROTOCOL.get(file_system, file_system)
                store_params_with_root = store_params.copy()
                if fs_protocol == "file" and not is_absolute_path(root):
                    store_params_with_root["root"] = f"{base_dir}/{root}"
                else:
                    store_params_with_root["root"] = root
                # See if there already is a store with this configuration
                data_store_config = DataStoreConfig(
                    store_id=fs_protocol, store_params=store_params_with_root
                )
                store_instance_id = data_store_pool.get_store_instance_id(
                    data_store_config
                )
                if not store_instance_id:
                    # Create new store with new unique store instance id
                    counter = 1
                    while data_store_pool.has_store_instance(
                        f"{fs_protocol}_{counter}"
                    ):
                        counter += 1
                    store_instance_id = f"{fs_protocol}_{counter}"
                    data_store_pool.add_store_config(
                        store_instance_id, data_store_config
                    )
                for config in config_list:
                    path = config["Path"]
                    if path.startswith(root) and config.get("StoreInstanceId") is None:
                        config["StoreInstanceId"] = store_instance_id
                        new_path = path[len(root) :]
                        while new_path.startswith("/") or new_path.startswith("\\"):
                            new_path = new_path[1:]
                        config["Path"] = new_path

    @classmethod
    def _get_other_store_params_than_root(cls, dataset_config: DatasetConfig) -> dict:
        file_system = FS_TYPE_TO_PROTOCOL.get(dataset_config.get("FileSystem", "file"))
        if file_system != "s3":
            return {}
        storage_options = dict()
        if "Anonymous" in dataset_config:
            storage_options["anon"] = dataset_config["Anonymous"]
        client_kwargs = dict()
        if "Endpoint" in dataset_config:
            client_kwargs["endpoint_url"] = dataset_config["Endpoint"]
        if "Region" in dataset_config:
            client_kwargs["region_name"] = dataset_config["Region"]
        storage_options["client_kwargs"] = client_kwargs
        store_params = dict(storage_options=storage_options)
        return store_params

    @classmethod
    def get_dataset_configs_from_stores(
        cls, data_store_pool: DataStorePool
    ) -> list[DatasetConfig]:
        all_dataset_configs: list[DatasetConfig] = []
        for store_instance_id in data_store_pool.store_instance_ids:
            LOG.info(f"Scanning store {store_instance_id!r}")
            data_store_config = data_store_pool.get_store_config(store_instance_id)

            # Note by forman: This iterator chaining is inefficient.
            # Preferably, we should offer
            #
            # store_dataset_ids = data_store.get_data_ids(
            #     data_type=(DATASET_TYPE, MULTI_LEVEL_DATASET_TYPE)
            # )
            #

            store_dataset_configs: list[ServerConfig] = data_store_config.user_data
            if store_dataset_configs:
                for store_dataset_config in store_dataset_configs:
                    dataset_id_pattern = store_dataset_config.get("Path", "*")
                    if _is_wildcard(dataset_id_pattern):
                        if store_instance_id in DATA_STORE_IDS_WARNING:
                            warnings.warn(
                                f"The data store with ID '{store_instance_id}' has "
                                "many data IDs. Using wildcard patterns to select "
                                "datasets may cause a long setup time of the server."
                            )
                        data_store = data_store_pool.get_store(store_instance_id)
                        store_dataset_ids = itertools.chain(
                            data_store.get_data_ids(data_type=DATASET_TYPE),
                            data_store.get_data_ids(data_type=MULTI_LEVEL_DATASET_TYPE),
                        )
                        for store_dataset_id in store_dataset_ids:
                            if fnmatch.fnmatch(store_dataset_id, dataset_id_pattern):
                                all_dataset_configs.append(
                                    _get_selected_dataset_config(
                                        store_dataset_id,
                                        store_instance_id,
                                        store_dataset_config,
                                    )
                                )
                    else:
                        all_dataset_configs.append(
                            _get_selected_dataset_config(
                                store_dataset_config["Path"],
                                store_instance_id,
                                store_dataset_config,
                            )
                        )

        # # Just for testing:
        # debug_file = 'all_dataset_configs.json'
        # with open(debug_file, 'w') as stream:
        #     json.dump(all_dataset_configs, stream)
        #     LOG.debug(f'Wrote file {debug_file!r}')

        return all_dataset_configs

    def new_dataset_metadata(
        self, store_instance_id: str, dataset_id: str
    ) -> Optional[DatasetDescriptor]:
        data_store = self._data_store_pool.get_store(store_instance_id)
        dataset_metadata = data_store.describe_data(dataset_id, data_type="dataset")
        if dataset_metadata.crs is not None:
            crs = pyproj.CRS.from_string(dataset_metadata.crs)
            if not crs.is_geographic:
                LOG.warning(
                    f"Ignoring dataset {dataset_id!r} from"
                    f" store instance {store_instance_id!r}"
                    f" because it uses a non-geographic CRS"
                )
                return None
        # noinspection PyTypeChecker
        return dataset_metadata

    def get_dataset_config(self, ds_id: str) -> DatasetConfig:
        dataset_configs = self.get_dataset_configs()
        dataset_config = next(
            (dsd for dsd in dataset_configs if dsd["Identifier"] == ds_id), None
        )
        if dataset_config is None:
            raise ApiError.NotFound(f'Dataset "{ds_id}" not found')
        return dataset_config

    def get_dataset_configs(self) -> list[DatasetConfig]:
        assert self._dataset_configs is not None
        return self._dataset_configs

    def get_dataset_groups_configs(self) -> list[DatasetConfig]:
        assert self._dataset_groups_config is not None
        return self._dataset_groups_config

    def get_entrypoint_dataset_id(self) -> str | None:
        if self.entrypoint_dataset_id:
            return self.entrypoint_dataset_id
        return None

    def get_data_store_pool(self) -> DataStorePool:
        assert self._data_store_pool is not None
        return self._data_store_pool

    @classmethod
    def _process_dataset_configs(
        cls, config: ServerConfig, base_dir: str
    ) -> tuple[DataStorePool, list[dict[str, Any]], str, list[dict[str, Any]]]:
        data_store_configs = config.get("DataStores", [])
        dataset_configs = config.get("Datasets", [])
        entrypoint_dataset_id = config.get("EntrypointDatasetId", "")
        dataset_groups_configs = config.get("DatasetGroups", [])

        data_store_pool = DataStorePool()
        for data_store_config_dict in data_store_configs:
            store_instance_id = data_store_config_dict.get("Identifier")
            store_id = data_store_config_dict.get("StoreId")
            store_params = data_store_config_dict.get("StoreParams", {})
            store_dataset_configs = data_store_config_dict.get("Datasets")
            store_config = DataStoreConfig(
                store_id, store_params=store_params, user_data=store_dataset_configs
            )
            data_store_pool.add_store_config(store_instance_id, store_config)
        dataset_configs = dataset_configs + cls.get_dataset_configs_from_stores(
            data_store_pool
        )
        # Allow dataset_configs to be writable, because
        # _maybe_assign_store_instance_ids() will change
        # entries:
        dataset_configs = [dict(c) for c in dataset_configs]
        cls._maybe_assign_store_instance_ids(dataset_configs, data_store_pool, base_dir)
        return (
            data_store_pool,
            dataset_configs,
            entrypoint_dataset_id,
            dataset_groups_configs,
        )

    def get_rgb_color_mapping(
        self, ds_id: str, norm_range: tuple[float, float] = (0.0, 1.0)
    ) -> tuple[list[Optional[str]], list[tuple[float, float]]]:
        var_names = [None, None, None]
        norm_ranges = [norm_range, norm_range, norm_range]
        color_mappings = self.get_color_mappings(ds_id)
        if color_mappings:
            rgb_mapping = color_mappings.get("rgb")
            if rgb_mapping:
                components = "Red", "Green", "Blue"
                for i in range(3):
                    component = components[i]
                    component_config = rgb_mapping.get(component, {})
                    var_name = component_config.get("Variable")
                    norm_vmin, norm_vmax = component_config.get(
                        "ValueRange", norm_range
                    )
                    var_names[i] = var_name
                    norm_ranges[i] = norm_vmin, norm_vmax
        return var_names, norm_ranges

    def get_color_mapping(
        self, ds_id: str, var_name: str
    ) -> tuple[str, str, tuple[float, float]]:
        if split_var_assignment(var_name)[1]:
            # var_name is an assignment expression, the result is unknown yet,
            # so we must return default values here.
            return DEFAULT_CMAP_NAME, DEFAULT_CMAP_NORM, DEFAULT_VALUE_RANGE

        cmap_name = None
        cmap_norm = None
        cmap_vmin, cmap_vmax = None, None
        color_mappings = self.get_color_mappings(ds_id)
        if color_mappings:
            color_mapping = color_mappings.get(var_name)
            if color_mapping:
                assert "ColorFile" not in color_mapping
                cmap_name = color_mapping.get("ColorBar")
                cmap_norm = color_mapping.get("Norm")
                cmap_vmin, cmap_vmax = color_mapping.get("ValueRange", (None, None))

        cmap_range = cmap_vmin, cmap_vmax
        if cmap_name and cmap_norm and None not in cmap_range:
            # noinspection PyTypeChecker
            return cmap_name, cmap_norm, cmap_range

        ds = self.get_dataset(ds_id, expected_var_names=[var_name])
        var = ds[var_name]
        valid_range = get_var_valid_range(var)
        return get_var_cmap_params(var, cmap_name, cmap_norm, cmap_range, valid_range)

    def _get_cm_styles(self) -> tuple[dict[str, Any], ColormapRegistry]:
        custom_colormaps = {}
        cm_styles = {}

        # go through CustomColorMaps
        for ctx_cmap in self.config.get("CustomColorMaps", []):
            custom_colormap, config_parse = create_colormap_from_config(ctx_cmap)
            custom_colormaps[custom_colormap.cm_name] = custom_colormap
        # go through Styles
        for style in self.config.get("Styles", []):
            style_id = style["Identifier"]
            color_mappings = dict()
            for var_name, color_mapping in style["ColorMappings"].items():
                if "ColorFile" in color_mapping:
                    custom_cmap_path = self.get_config_path(
                        color_mapping, "ColorMappings", path_entry_name="ColorFile"
                    )
                    custom_colormap = custom_colormaps.get(custom_cmap_path)
                    if custom_colormap is None:
                        custom_colormap = load_custom_colormap(custom_cmap_path)
                        if custom_colormap is not None:
                            custom_colormaps[custom_cmap_path] = custom_colormap
                    if custom_colormap is not None:
                        color_mappings[var_name] = {
                            "ColorBar": custom_colormap.cm_name,
                            "ValueRange": (
                                custom_colormap.norm.vmin,
                                custom_colormap.norm.vmax,
                            ),
                        }
                else:
                    if color_mapping.get("ColorBar") in custom_colormaps:
                        cmap = custom_colormaps[color_mapping["ColorBar"]]
                        if (
                            cmap.cm_type == "categorical"
                            and "ValueRange" in color_mapping
                        ):
                            LOG.warning(
                                f"Custom color map {color_mapping['ColorBar']!r} is "
                                "categorical. ValueRange is ignored."
                            )
                            color_mapping = dict(
                                ColorBar=color_mapping["ColorBar"],
                                ValueRange=[
                                    min(cmap.values),
                                    max(cmap.values),
                                ],
                            )
                        if "ValueRange" not in color_mapping:
                            color_mapping = dict(
                                ColorBar=color_mapping["ColorBar"],
                                ValueRange=[
                                    min(cmap.values),
                                    max(cmap.values),
                                ],
                            )
                    color_mappings[var_name] = dict(color_mapping)

            cm_styles[style_id] = color_mappings

        return cm_styles, ColormapRegistry(*custom_colormaps.values())

    def get_color_mappings(self, ds_id: str) -> Optional[dict[str, dict[str, Any]]]:
        dataset_config = self.get_dataset_config(ds_id)
        style_id = dataset_config.get("Style", "default")
        return self._cm_styles.get(style_id, {})

    def _get_dataset_entry(self, ds_id: str) -> tuple[MultiLevelDataset, ServerConfig]:
        if ds_id not in self._dataset_cache:
            with self.rlock:
                self._set_dataset_entry(self._create_dataset_entry(ds_id))
        return self._dataset_cache[ds_id]

    def _set_dataset_entry(
        self, dataset_entry: tuple[MultiLevelDataset, DatasetConfig]
    ):
        ml_dataset, dataset_config = dataset_entry
        self._dataset_cache[ml_dataset.ds_id] = ml_dataset, dataset_config

    def _create_dataset_entry(
        self, ds_id: str
    ) -> tuple[MultiLevelDataset, DatasetConfig]:
        dataset_config = self.get_dataset_config(ds_id)
        ml_dataset = self._open_ml_dataset(dataset_config)
        return ml_dataset, dataset_config

    def _open_ml_dataset(self, dataset_config: DatasetConfig) -> MultiLevelDataset:
        ds_id: str = dataset_config.get("Identifier")
        store_instance_id = dataset_config.get("StoreInstanceId")
        if store_instance_id:
            data_store_pool = self.get_data_store_pool()
            data_store = data_store_pool.get_store(store_instance_id)
            data_id = dataset_config.get("Path")
            open_params = dict(dataset_config.get("StoreOpenParams") or {})
            open_params_schema = data_store.get_open_data_params_schema(data_id=data_id)

            # Inject cache_size=chunk_cache_capacity, if given and possible
            chunk_cache_capacity = self.get_dataset_chunk_cache_capacity(dataset_config)
            if (
                chunk_cache_capacity
                and "cache_size" not in open_params
                and "cache_size" in open_params_schema.properties
            ):
                open_params["cache_size"] = chunk_cache_capacity

            # Inject data_type="mldataset", if possible
            if (
                "data_type" not in open_params
                and "data_type" in open_params_schema.properties
                and "mldataset" in data_store.get_data_types()
            ):
                open_params["data_type"] = "mldataset"

            with self.measure_time(
                tag=f"Opened dataset {ds_id!r} from data store {store_instance_id!r}"
            ):
                dataset = data_store.open_data(data_id, **open_params)

            if isinstance(dataset, MultiLevelDataset):
                # Expected, nominal case.
                ml_dataset: MultiLevelDataset = dataset
            else:
                # Fallback. Usually results in poor tile computation performance.
                ml_dataset = BaseMultiLevelDataset(dataset)

            ml_dataset.ds_id = ds_id

        else:
            fs_type = dataset_config.get("FileSystem")
            if fs_type != "memory":
                raise ApiError.InvalidServerConfig(
                    f"Invalid FileSystem {fs_type!r} in dataset configuration {ds_id!r}"
                )

            with self.measure_time(tag=f"Opened dataset {ds_id!r} from {fs_type!r}"):
                ml_dataset = _open_ml_dataset_from_python_code(self, dataset_config)

        augmentation = dataset_config.get("Augmentation")
        if augmentation:
            script_path = self.get_config_path(
                augmentation, f"'Augmentation' of dataset configuration {ds_id}"
            )
            input_parameters = augmentation.get("InputParameters", {})
            input_parameters = self.eval_config_value(input_parameters)
            callable_name = augmentation.get("Class")
            is_factory = callable_name is not None
            if not is_factory:
                callable_name = augmentation.get("Function", COMPUTE_VARIABLES)
            ml_dataset = augment_ml_dataset(
                ml_dataset,
                script_path,
                callable_name,
                self.get_ml_dataset,
                self.set_ml_dataset,
                input_parameters=input_parameters,
                is_factory=is_factory,
                exception_type=ApiError.InvalidServerConfig,
            )
        return ml_dataset

    def get_legend_label(self, ds_id: str, var_name: str):
        dataset = self.get_dataset(ds_id)
        if var_name in dataset:
            ds = self.get_dataset(ds_id)
            units = ds[var_name].units
            return units
        raise ApiError.NotFound(f'Variable "{var_name}" not found in dataset "{ds_id}"')

    def get_dataset_place_groups(
        self, ds_id: str, base_url: str, load_features=False
    ) -> list[dict]:
        dataset_config = self.get_dataset_config(ds_id)

        place_group_id_prefix = f"DS-{ds_id}-"

        def predicate(pg_id: str, _: Any):
            return pg_id.startswith(place_group_id_prefix)

        place_groups = self.places_ctx.get_cached_place_groups(predicate=predicate)
        if place_groups:
            # for a given dataset, all place groups cached,
            # so the returned list is complete
            return place_groups

        place_groups = self.places_ctx.load_place_groups(
            dataset_config.get("PlaceGroups", []),
            base_url,
            is_global=False,
            load_features=load_features,
            qualifiers=[ds_id],
        )

        for place_group in place_groups:
            place_group_id = place_group_id_prefix + place_group["id"]
            self.places_ctx.set_cached_place_group(place_group_id, place_group)

        return place_groups

    def get_dataset_place_group(
        self, ds_id: str, place_group_id: str, base_url: str, load_features=False
    ) -> dict:
        place_groups = self.get_dataset_place_groups(
            ds_id, base_url, load_features=False
        )
        for place_group in place_groups:
            if place_group_id == place_group["id"]:
                if load_features:
                    self.places_ctx.load_place_group_features(place_group)
                return place_group
        raise ApiError.NotFound(f'Place group "{place_group_id}" not found')

    def get_dataset_and_coord_variable(self, ds_name: str, dim_name: str):
        ds = self.get_dataset(ds_name)
        if dim_name not in ds.coords:
            raise ApiError.NotFound(
                f"Dimension {dim_name!r} has no coordinates in dataset {ds_name!r}"
            )
        return ds, ds.coords[dim_name]

    @classmethod
    def get_var_indexers(
        cls,
        ds_name: str,
        var_name: str,
        var: xr.DataArray,
        dim_names: list[str],
        query_params: dict[str, str],
    ) -> dict[str, Any]:
        var_indexers = dict()
        for dim_name in dim_names:
            if dim_name not in var.coords:
                raise ApiError.BadRequest(
                    f"Dimension {dim_name!r} of variable {var_name!r}"
                    f" in dataset {ds_name!r} has no coordinates"
                )
            coord_var = var.coords[dim_name]
            dim_value_str = query_params.get(dim_name)
            try:
                if dim_value_str is None:
                    var_indexers[dim_name] = coord_var.values[0]
                elif dim_value_str == "current":
                    var_indexers[dim_name] = coord_var.values[-1]
                elif np.issubdtype(coord_var.dtype, np.floating):
                    var_indexers[dim_name] = float(dim_value_str)
                elif np.issubdtype(coord_var.dtype, np.integer):
                    var_indexers[dim_name] = int(dim_value_str)
                elif np.issubdtype(coord_var.dtype, np.datetime64):
                    if "/" in dim_value_str:
                        date_str_1, date_str_2 = dim_value_str.split("/", maxsplit=1)
                        var_indexer_1 = pd.to_datetime(date_str_1)
                        var_indexer_2 = pd.to_datetime(date_str_2)
                        var_indexers[dim_name] = (
                            var_indexer_1 + (var_indexer_2 - var_indexer_1) / 2
                        )
                    else:
                        date_str = dim_value_str
                        var_indexers[dim_name] = pd.to_datetime(date_str)
                else:
                    raise ValueError(
                        f"unable to convert value"
                        f" {dim_value_str!r} to {coord_var.dtype!r}"
                    )
            except ValueError as e:
                raise ApiError.BadRequest(
                    f"{dim_value_str!r} is not a valid value"
                    f" for dimension {dim_name!r} "
                    f"of variable {var_name!r} of dataset {ds_name!r}"
                ) from e
        return var_indexers

    def get_dataset_chunk_cache_capacity(
        self, dataset_config: DatasetConfig
    ) -> Optional[int]:
        cache_size = self.get_chunk_cache_capacity(dataset_config, "ChunkCacheSize")
        if cache_size is None:
            cache_size = self.get_chunk_cache_capacity(
                self.config, "DatasetChunkCacheSize"
            )
        return cache_size

    @classmethod
    def get_chunk_cache_capacity(
        cls, config: Mapping[str, Any], cache_size_key: str
    ) -> Optional[int]:
        cache_size = config.get(cache_size_key, None)
        if not cache_size:
            return None
        elif isinstance(cache_size, str):
            try:
                cache_size = parse_mem_size(cache_size)
            except ValueError:
                raise ApiError.InvalidServerConfig(f"Invalid {cache_size_key}")
        elif not isinstance(cache_size, int) or cache_size < 0:
            raise ApiError.InvalidServerConfig(f"Invalid {cache_size_key}")
        return cache_size


def _open_ml_dataset_from_python_code(
    ctx: DatasetsContext, dataset_config: DatasetConfig
) -> MultiLevelDataset:
    ds_id = dataset_config.get("Identifier")
    path = ctx.get_config_path(dataset_config, f"dataset configuration {ds_id}")
    callable_name = dataset_config.get("Class")
    is_factory = callable_name is not None
    if not is_factory:
        callable_name = dataset_config.get("Function", COMPUTE_DATASET)
    input_dataset_ids = dataset_config.get("InputDatasets", [])
    input_parameters = dataset_config.get("InputParameters", {})
    input_parameters = ctx.eval_config_value(input_parameters)
    chunk_cache_capacity = ctx.get_dataset_chunk_cache_capacity(dataset_config)
    if chunk_cache_capacity:
        warnings.warn(
            "chunk cache size is not effective for datasets computed from scripts"
        )
    for input_dataset_id in input_dataset_ids:
        if not ctx.get_dataset_config(input_dataset_id):
            raise ApiError.InvalidServerConfig(
                f"Invalid dataset configuration {ds_id!r}:"
                f" Input dataset {input_dataset_id!r} of"
                f" callable {callable_name!r}"
                f" must reference another dataset"
            )
    return open_ml_dataset_from_python_code(
        path,
        callable_name=callable_name,
        input_ml_dataset_ids=input_dataset_ids,
        input_ml_dataset_getter=ctx.get_ml_dataset,
        input_parameters=input_parameters,
        is_factory=is_factory,
        ds_id=ds_id,
        exception_type=ApiError.InvalidServerConfig,
    )


def _is_not_empty(prefix):
    return prefix != "" and prefix != "/" and prefix != "\\"


def _get_roots(paths):
    list.sort(paths)
    prefixes = _get_common_prefixes(paths)
    if len(prefixes) < 1:
        roots = [""]
    else:
        # perform further step to merge prefixes with same start
        prefixes = list(set(prefixes))
        prefixes.sort()
        roots = []
        root_candidate = prefixes[0]
        for root in prefixes[1:]:
            common_root = os.path.commonprefix([root_candidate, root])
            if _is_not_empty(common_root):
                root_candidate = common_root
            else:
                roots.append(root_candidate)
                root_candidate = root
        roots.append(root_candidate)
    if roots[0] == "":
        roots.append(roots.pop(0))
    return roots


def _get_common_prefixes(p):
    # Recursively examine a list of paths for common prefixes:
    # If no common prefix is found, split the list in half and
    # examine each half separately
    prefix = os.path.commonprefix(p)
    # ensure prefix does not end with full or partial directory
    # or file name
    prefix = prefix[: max(_last_index(prefix, "/"), _last_index(prefix, "\\"), 0)]
    if _is_not_empty(prefix) or len(p) == 1:
        return [prefix]
    else:
        return _get_common_prefixes(p[: int(len(p) / 2)]) + _get_common_prefixes(
            p[int(len(p) / 2) :]
        )


def _get_selected_dataset_config(
    store_dataset_id: str, store_instance_id: str, dataset_config_base: dict
) -> dict:
    LOG.debug(f"Selected dataset {store_dataset_id!r}")
    dataset_config = dict(StoreInstanceId=store_instance_id, **dataset_config_base)
    if "Identifier" in dataset_config and dataset_config["Path"] != store_dataset_id:
        raise ApiError.InvalidServerConfig(
            "User-defined identifiers can only be assigned"
            " to datasets with non-wildcard paths."
        )
    elif "Identifier" not in dataset_config:
        dataset_config["Path"] = store_dataset_id
        dataset_config["Identifier"] = (
            f"{store_instance_id}{STORE_DS_ID_SEPARATOR}{store_dataset_id}"
        )
    return dataset_config


def _last_index(prefix, symbol):
    try:
        return prefix.rindex(symbol)
    except ValueError:
        return -1


_MULTI_LEVEL_DATASET_OPENERS = {
    "memory": _open_ml_dataset_from_python_code,
}


def _is_wildcard(string: str) -> bool:
    return "?" in string or "*" in string


def _new_dataset_config(**kwargs) -> dict[str, Any]:
    return {k: v for k, v in kwargs.items() if v is not None}



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import click

DEFAULT_OUTPUT_PATH = "{input}-optimized.zarr"


# noinspection PyShadowingBuiltins
@click.command(name="optimize")
@click.argument("CUBE")
@click.option(
    "--output",
    "-o",
    metavar="OUTPUT",
    help=f'Output path. The placeholder "{input}" will be replaced by the input\'s filename '
    f'without extension (such as ".zarr"). Defaults to "{DEFAULT_OUTPUT_PATH}".',
    default=DEFAULT_OUTPUT_PATH,
)
@click.option(
    "--in-place",
    "-I",
    "in_place",
    help="Optimize cube in place. Ignores output path.",
    is_flag=True,
)
@click.option(
    "--coords",
    "-C",
    "unchunk_coords",
    help="Also optimize coordinate variables by converting any chunked arrays "
    "into single, non-chunked, contiguous arrays.",
    is_flag=True,
)
def optimize(cube, output, in_place, unchunk_coords):
    """
    Optimize xcube dataset for faster access.

    Reduces the number of metadata and coordinate data files in xcube dataset given by CUBE.
    Consolidated cubes open much faster especially from remote locations, e.g. in object storage,
    because obviously much less HTTP requests are required to fetch initial cube meta
    information. That is, it merges all metadata files into a single top-level JSON file ".zmetadata".
    Optionally, it removes any chunking of coordinate variables
    so they comprise a single binary data file instead of one file per data chunk.
    The primary usage of this command is to optimize data cubes for cloud object storage.
    The command currently works only for data cubes using ZARR format.
    """
    from xcube.core.optimize import optimize_dataset

    optimize_dataset(
        cube, output_path=output, in_place=in_place, unchunk_coords=unchunk_coords
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os.path
import uuid
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Optional

import xarray as xr

from xcube.core.byoa import CodeConfig, FileSet
from xcube.core.gridmapping import GridMapping
from xcube.util.assertions import assert_given, assert_instance, assert_true
from xcube.util.perf import measure_time

from .abc import MultiLevelDataset
from .lazy import LazyMultiLevelDataset

MultiLevelDatasetGetter = Callable[[str], MultiLevelDataset]
MultiLevelDatasetSetter = Callable[[MultiLevelDataset], None]


class ComputedMultiLevelDataset(LazyMultiLevelDataset):
    """A multi-level dataset whose level datasets are computed
    by a user function.

    The script can import other Python modules located in the same
    directory as *script_path*.
    """

    def __init__(
        self,
        script_path: str,
        callable_name: str,
        input_ml_dataset_ids: Sequence[str],
        input_ml_dataset_getter: MultiLevelDatasetGetter,
        input_parameters: Optional[Mapping[str, Any]] = None,
        ds_id: str = "",
        exception_type: type = ValueError,
    ):
        callable_ref, callable_obj = self.get_callable(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            ds_id=ds_id,
            exception_type=exception_type,
        )

        super().__init__(ds_id=ds_id, parameters=input_parameters)
        self._callable_ref = callable_ref
        self._callable_obj = callable_obj
        self._input_ml_dataset_ids = input_ml_dataset_ids
        self._input_ml_dataset_getter = input_ml_dataset_getter
        self._exception_type = exception_type

    @classmethod
    def get_callable(
        cls,
        script_path: str,
        callable_name: str,
        input_ml_dataset_ids: Sequence[str],
        input_ml_dataset_getter: MultiLevelDatasetGetter,
        input_parameters: Optional[Mapping[str, Any]] = None,
        ds_id: str = "",
        exception_type: type = ValueError,
    ) -> tuple[str, Callable]:
        assert_instance(script_path, str, name="script_path")
        assert_given(script_path, name="script_path")
        assert_true(
            callable(input_ml_dataset_getter),
            message=f"input_ml_dataset_getter must be a callable",
        )
        assert_given(input_ml_dataset_getter, name="input_ml_dataset_getter")
        assert_instance(ds_id, str, name="ds_id")
        assert_given(ds_id, name="ds_id")

        module_name = None
        basename = os.path.basename(script_path)
        basename, ext = os.path.splitext(basename)
        if ext == ".py":
            script_path = os.path.dirname(script_path)
            module_name = basename

        if ":" in callable_name:
            callable_ref = callable_name
        else:
            if not module_name:
                raise exception_type(
                    f"Invalid in-memory dataset descriptor {ds_id!r}:"
                    f" Missing module name in {callable_name!r}"
                )
            callable_ref = f"{module_name}:{callable_name}"

        if not input_ml_dataset_ids:
            raise exception_type(
                f"Invalid in-memory dataset descriptor {ds_id!r}:"
                f" Input dataset(s) missing for callable {callable_name!r}"
            )

        for input_param_name in (input_parameters or {}).keys():
            if not input_param_name or not input_param_name.isidentifier():
                raise exception_type(
                    f"Invalid in-memory dataset descriptor {ds_id!r}:"
                    f" Input parameter {input_param_name!r}"
                    f" for callable {callable_name!r}"
                    f" is not a valid Python identifier"
                )

        try:
            callable_obj = CodeConfig.from_file_set(
                FileSet(path=script_path),
                callable_ref=callable_ref,
                install_required=False,
            ).get_callable()
        except (TypeError, ValueError, ImportError) as e:
            raise exception_type(f"Invalid dataset descriptor {ds_id!r}: {e}") from e

        return callable_ref, callable_obj

    @property
    def num_inputs(self) -> int:
        return len(self._input_ml_dataset_ids)

    def get_input_dataset(self, index: int) -> MultiLevelDataset:
        return self._input_ml_dataset_getter(self._input_ml_dataset_ids[index])

    def _get_num_levels_lazily(self) -> int:
        return self.get_input_dataset(0).num_levels

    def _get_grid_mapping_lazily(self) -> GridMapping:
        return self.get_input_dataset(0).grid_mapping

    def _get_dataset_lazily(self, index: int, parameters: dict[str, Any]) -> xr.Dataset:
        input_datasets = [
            self._input_ml_dataset_getter(ds_id).get_dataset(index)
            for ds_id in self._input_ml_dataset_ids
        ]
        try:
            with measure_time(
                f"Computed in-memory dataset {self.ds_id!r} at level {index}"
            ):
                computed_value = self._callable_obj(*input_datasets, **parameters)
        except Exception as e:
            raise self._exception_type(
                f"Failed to compute in-memory dataset {self.ds_id!r}"
                f" at level {index} "
                f"from function {self._callable_ref!r}(): {e}"
            ) from e
        if not isinstance(computed_value, xr.Dataset):
            raise self._exception_type(
                f"Failed to compute in-memory dataset {self.ds_id!r}"
                f" at level {index} "
                f"from function {self._callable_ref!r}(): "
                f"expected an xarray.Dataset but got {type(computed_value)}"
            )
        return computed_value


def augment_ml_dataset(
    ml_dataset: MultiLevelDataset,
    script_path: str,
    callable_name: str,
    input_ml_dataset_getter: MultiLevelDatasetGetter,
    input_ml_dataset_setter: MultiLevelDatasetSetter,
    input_parameters: Optional[Mapping[str, Any]] = None,
    is_factory: bool = False,
    exception_type: type = ValueError,
):
    from .combined import CombinedMultiLevelDataset
    from .identity import IdentityMultiLevelDataset

    with measure_time(f"Added augmentation from {script_path}"):
        orig_id = ml_dataset.ds_id
        aug_id = uuid.uuid4()
        aug_inp_id = f"aug-input-{aug_id}"
        aug_inp_ds = IdentityMultiLevelDataset(ml_dataset, ds_id=aug_inp_id)
        input_ml_dataset_setter(aug_inp_ds)
        aug_ds = _open_ml_dataset_from_python_code(
            script_path,
            callable_name,
            [aug_inp_id],
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            is_factory=is_factory,
            ds_id=f"aug-{aug_id}",
            exception_type=exception_type,
        )
        return CombinedMultiLevelDataset([ml_dataset, aug_ds], ds_id=orig_id)


def open_ml_dataset_from_python_code(
    script_path: str,
    callable_name: str,
    input_ml_dataset_ids: Sequence[str],
    input_ml_dataset_getter: MultiLevelDatasetGetter,
    input_parameters: Optional[Mapping[str, Any]] = None,
    is_factory: bool = False,
    ds_id: str = "",
    exception_type: type = ValueError,
) -> MultiLevelDataset:
    with measure_time(f"Opened memory dataset {script_path}"):
        return _open_ml_dataset_from_python_code(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            is_factory=is_factory,
            ds_id=ds_id,
            exception_type=exception_type,
        )


def _open_ml_dataset_from_python_code(
    script_path: str,
    callable_name: str,
    input_ml_dataset_ids: Sequence[str],
    input_ml_dataset_getter: MultiLevelDatasetGetter,
    input_parameters: Optional[Mapping[str, Any]] = None,
    is_factory: bool = False,
    ds_id: str = "",
    exception_type: type = ValueError,
) -> MultiLevelDataset:
    if is_factory:
        callable_ref, callable_obj = ComputedMultiLevelDataset.get_callable(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            ds_id=ds_id,
            exception_type=exception_type,
        )
        input_datasets = [
            input_ml_dataset_getter(ds_id) for ds_id in input_ml_dataset_ids
        ]
        try:
            ml_dataset = callable_obj(*input_datasets, **(input_parameters or {}))
            if not isinstance(ml_dataset, MultiLevelDataset):
                raise TypeError(
                    f"{callable_ref!r} must return instance of"
                    f" xcube.core.mldataset.MultiLevelDataset,"
                    f" but was {type(ml_dataset)}"
                )
            ml_dataset.ds_id = ds_id
            return ml_dataset
        except BaseException as e:
            raise exception_type(
                f"Invalid in-memory dataset descriptor {ds_id!r}: {e}"
            ) from e
    else:
        return ComputedMultiLevelDataset(
            script_path,
            callable_name,
            input_ml_dataset_ids,
            input_ml_dataset_getter,
            input_parameters=input_parameters,
            ds_id=ds_id,
            exception_type=exception_type,
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import json
import os.path
from collections.abc import Sequence
from typing import List

import numpy as np
import xarray as xr
import zarr


def unchunk_dataset(
    dataset_path: str, var_names: Sequence[str] = None, coords_only: bool = False
):
    """Unchunk dataset variables in-place.

    Args:
        dataset_path: Path to ZARR dataset directory.
        var_names: Optional list of variable names.
        coords_only: Un-chunk coordinate variables only.
    """

    is_zarr = os.path.isfile(os.path.join(dataset_path, ".zgroup"))
    if not is_zarr:
        raise ValueError(f"{dataset_path!r} is not a valid Zarr directory")

    with xr.open_zarr(dataset_path) as dataset:
        if var_names is None:
            if coords_only:
                var_names = list(dataset.coords)
            else:
                var_names = list(dataset.variables)
        else:
            for var_name in var_names:
                if coords_only:
                    if var_name not in dataset.coords:
                        raise ValueError(
                            f"variable {var_name!r} is not a coordinate variable in {dataset_path!r}"
                        )
                else:
                    if var_name not in dataset.variables:
                        raise ValueError(
                            f"variable {var_name!r} is not a variable in {dataset_path!r}"
                        )

    _unchunk_vars(dataset_path, var_names)


def _unchunk_vars(dataset_path: str, var_names: list[str]):
    for var_name in var_names:
        var_path = os.path.join(dataset_path, var_name)

        # Optimization: if "shape" and "chunks" are equal in ${var}/.zarray, we are done
        var_array_info_path = os.path.join(var_path, ".zarray")
        with open(var_array_info_path) as fp:
            var_array_info = json.load(fp)
            if var_array_info.get("shape") == var_array_info.get("chunks"):
                continue

        # Open array and remove chunks from the data
        var_array = zarr.convenience.open_array(var_path, "r+")
        if var_array.shape != var_array.chunks:
            # TODO (forman): Fully loading data is inefficient and dangerous for large arrays.
            #                Instead save unchunked to temp and replace existing chunked array dir with temp.
            # Fully load data and attrs so we no longer depend on files
            data = np.array(var_array)
            attributes = var_array.attrs.asdict()
            # Save array data
            zarr.convenience.save_array(
                var_path, data, chunks=False, fill_value=var_array.fill_value
            )
            # zarr.convenience.save_array() does not seem save user attributes (file ".zattrs" not written),
            # therefore we must modify attrs explicitly:
            var_array = zarr.convenience.open_array(var_path, "r+")
            var_array.attrs.update(attributes)

    zarr.consolidate_metadata(dataset_path)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Any, Callable, Dict

import xarray as xr

from .abc import MultiLevelDataset
from .lazy import LazyMultiLevelDataset


class MappedMultiLevelDataset(LazyMultiLevelDataset):
    def __init__(
        self,
        ml_dataset: MultiLevelDataset,
        mapper_function: Callable[[xr.Dataset], xr.Dataset],
        ds_id: str = None,
        mapper_params: dict[str, Any] = None,
    ):
        """"""
        super().__init__(ds_id=ds_id, parameters=mapper_params)
        self._ml_dataset = ml_dataset
        self._mapper_function = mapper_function

    def _get_num_levels_lazily(self) -> int:
        return self._ml_dataset.num_levels

    def _get_dataset_lazily(
        self, index: int, mapper_params: dict[str, Any]
    ) -> xr.Dataset:
        return self._mapper_function(
            self._ml_dataset.get_dataset(index), **mapper_params
        )

    def close(self):
        self._ml_dataset.close()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.


class DataStoreError(Exception):
    """Raised on error in any of the data store, opener, or writer methods.

    Args:
        message: The error message.
    """

    def __init__(self, message: str):
        super().__init__(message)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.
from typing import Tuple

import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.normalize import encode_cube
from xcube.core.store import DataStorePool, get_data_store_instance, new_data_writer
from xcube.util.progress import observe_dask_progress

from ..config import OutputConfig


class CubeWriter:
    def __init__(self, output_config: OutputConfig, store_pool: DataStorePool = None):
        self._output_config = output_config
        self._store_pool = store_pool

    def write_cube(self, cube: xr.Dataset, gm: GridMapping) -> tuple[str, xr.Dataset]:
        output_config = self._output_config
        dataset = encode_cube(cube, grid_mapping=gm)
        with observe_dask_progress("writing cube", 100):
            write_params = output_config.write_params or {}
            store_params = output_config.store_params or {}
            if output_config.store_id:
                store_instance = get_data_store_instance(
                    output_config.store_id,
                    store_params=store_params,
                    store_pool=self._store_pool,
                )
                writer = store_instance.store
                write_params.update(writer_id=output_config.writer_id, **write_params)
            else:
                writer = new_data_writer(output_config.writer_id)
                write_params.update(**store_params, **write_params)
            if not dataset.attrs.get("title"):
                # Set fallback title, so we can distinguish
                # datasets from stores in xcube-viewer
                dataset = dataset.assign_attrs(title=output_config.data_id)
            data_id = writer.write_data(
                dataset,
                data_id=output_config.data_id,
                replace=output_config.replace or False,
                **write_params,
            )
        return data_id, dataset



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from typing import Dict, Set

import xarray as xr

ANCILLARY_VAR_NAME_PREFIXES_TO_STANDARD_NAME_MODIFIERS = [
    ("std", "standard_error"),
    ("count", "number_of_observations"),
]


def find_ancillary_var_names(
    dataset: xr.Dataset,
    var_name: str,
    same_shape: bool = False,
    same_dims: bool = False,
) -> dict[str, set[str]]:
    variable = dataset.data_vars.get(var_name)

    results = {}

    # Check for CF compatibility according to
    # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#ancillary-data
    #
    if variable is not None and "ancillary_variables" in variable.attrs:
        ancillary_var_names = variable.attrs["ancillary_variables"].split(" ")
        for ancillary_var_name in ancillary_var_names:
            if ancillary_var_name in dataset.data_vars:
                ancillary_var = dataset.data_vars[ancillary_var_name]
                if (not same_shape or variable.shape == ancillary_var.shape) and (
                    not same_dims or variable.dims == variable.dims
                ):
                    standard_name_modifier = (
                        _get_standard_name_modifier(variable, ancillary_var) or ""
                    )
                    if standard_name_modifier not in results:
                        results[standard_name_modifier] = set()
                    results[standard_name_modifier].add(ancillary_var_name)

    if variable is not None and not results:
        #
        # Check for less strict CF compatibility (missing attribute 'ancillary_variables')
        #
        if "standard_name" in variable.attrs:
            for ancillary_var_name, ancillary_var in dataset.data_vars.items():
                if ancillary_var is variable:
                    continue
                standard_name_modifier = _get_standard_name_modifier(
                    variable, ancillary_var
                )
                if standard_name_modifier is not None:
                    if standard_name_modifier not in results:
                        results[standard_name_modifier] = set()
                    results[standard_name_modifier].add(ancillary_var_name)

    if not results:
        #
        # Search for variables with xcube-specific prefixes that indicate uncertainty:
        #
        for (
            prefix,
            standard_name_modifier,
        ) in ANCILLARY_VAR_NAME_PREFIXES_TO_STANDARD_NAME_MODIFIERS:
            ancillary_var_name = f"{var_name}_{prefix}"
            if ancillary_var_name in dataset.data_vars:
                if standard_name_modifier not in results:
                    results[standard_name_modifier] = set()
                results[standard_name_modifier].add(ancillary_var_name)

    return results


def _get_standard_name_modifier(variable, ancillary_var):
    """See CF Conventions v 1.7, Appendix C: Standard Name Modifiers"""
    if "standard_name" in ancillary_var.attrs:
        ancillary_var_std_name = ancillary_var.attrs["standard_name"]
        ancillary_var_std_name_parts = ancillary_var_std_name.split(" ")
        if len(ancillary_var_std_name_parts) == 2 and ancillary_var_std_name_parts[
            0
        ] == variable.attrs.get("standard_name"):
            return ancillary_var_std_name_parts[-1]
    return None
